import os.path
import numpy as np
import pandas as pd
import pickle
import astropy.units as u
from solarsystemMB import SSObject
from MESSENGERuvvs import MESSENGERdata
from .ModelResults import (ModelResult, read_format, results_loadfile,
                           results_packet_weighting)
from .database_connect import database_connect


class LOSResult(ModelResult):
    def __init__(self, inputs, data, quantity, dphi=3*u.deg,
                 filenames=None, overwrite=False):

        self.type = 'LineOfSight'
        self.species = inputs.options.atom
        self.quantity = quantity # column, radiance
        self.origin = inputs.geometry.planet
        self.unit = u.def_unit('R_' + self.origin.object,
                               self.origin.radius)
        self.dphi = dphi

        ModelResult.__init__(self, inputs)
        if isinstance(filenames, str):
            print('Setting filenames breaks calibration.')
            self.filenames = [filenames]
        elif isinstance(filenames, list):
            print('Setting filenames breaks calibration.')
            self.filenames = filenames
        else:
            pass

        if self.quantity == 'radiance':
            self.mechanism = 'resscat',

            if inputs.options.atom == 'Na':
                self.wavelength = 5891*u.AA, 5897*u.AA
            elif inputs.options.atom == 'Ca':
                self.wavelength = 4227*u.AA,
            elif inputs.options.atom == 'Mg':
                self.wavelength = 2852*u.AA,
            else:
                assert 0, f'No default wavelength for {input.options.atom}'
        else:
            pass

        nspec = len(data.x)
        self.radiance = np.zeros(nspec)
        self.ninview = np.zeros(nspec, dtype=int)

        for j,outfile in enumerate(self.filenames):
            # Search to see if it is already done
            radiance_, packets_, idnum = self.restore(data, outfile)

            if (radiance_ is None) or (overwrite):
                if (radiance_ is not None) and (overwrite):
                    self.delete_model(idnum)
                radiance_, packets_, = self.create_model(data, outfile)
                print(f'Completed model {j+1} of {len(self.filenames)}')
            else:
                print(f'Model {j+1} of {len(self.filenames)} '
                       'previously completed.')

            self.radiance += radiance_
            self.packets += packets_

        self.radiance = self.radiance * self.atoms_per_packet.value * u.R

    def delete_model(self, idnum):
        with database_connect() as con:
            cur = con.cursor()
            cur.execute('''SELECT idnum, filename FROM uvvsmodels
                           WHERE out_idnum = %s''', (idnum, ))
            for mid, mfile in cur.fetchall():
                cur.execute('''DELETE from uvvsmodels
                               WHERE idnum = %s''', (mid, ))
                if os.path.exists(mfile):
                    os.remove(mfile)

    def save(self, data, fname, radiance, packets):
        # Determine if the model can be saved.
        # Criteria: 1 complete orbit, nothing more.
        orbits = set(data.orbit)
        orb = orbits.pop()

        if len(orbits) != 0:
            print('Model spans more than one orbit. Cannot be saved.')
        else:
            mdata = MESSENGERdata(self.species, f'orbit = {orb}')
            if len(mdata) != len(data):
                print('Model does not contain the complete orbit. '
                      'Cannot be saved.')
            else:
                con = database_connect()
                cur = con.cursor()

                # Determine the id of the outputfile
                idnum_ = pd.read_sql(f'''SELECT idnum
                                        FROM outputfile
                                        WHERE filename='{fname}' ''', con)
                idnum = int(idnum_.idnum[0])

                # Insert the model into the database
                if self.quantity == 'radiance':
                    mech = ', '.join(sorted([m for m in self.mechanism]))
                    wave_ = sorted([w.value for w in self.wavelength])
                    wave = ', '.join([str(w) for w in wave_])
                else:
                    mech = None
                    wave = None

                cur.execute(f'''INSERT into uvvsmodels (out_idnum, quantity,
                                    orbit, dphi, mechanism, wavelength)
                                values (%s, %s, %s, %s, %s, %s)''',
                            (idnum, self.quantity, orb, self.dphi.value,
                             mech, wave))

                # Determine the savefile name
                idnum_ = pd.read_sql('''SELECT idnum
                                        FROM uvvsmodels
                                        WHERE filename is NULL''', con)
                assert len(idnum_) == 1
                idnum = int(idnum_.idnum[0])

                savefile = os.path.join(os.path.dirname(fname),
                                        f'model.orbit{orb:04}.{idnum}.pkl')
                with open(savefile, 'wb') as f:
                    pickle.dump((radiance, packets), f)
                cur.execute(f'''UPDATE uvvsmodels
                                SET filename=%s
                                WHERE idnum=%s''', (savefile, idnum))
                con.close()

    def restore(self, data, fname):
        # Determine if the model can be restored.
        # Criteria: 1 complete orbit, nothing more.
        orbits = set(data.orbit)
        orb = orbits.pop()

        if len(orbits) != 0:
            print('Model spans more than one orbit. Cannot be saved.')
            radiance, packets = None, None
        else:
            mdata = MESSENGERdata(self.species, f'orbit = {orb}')
            if len(mdata) != len(data):
                print('Model does not contain the complete orbit. '
                      'Cannot be saved.')
                radiance, packets, idnum = None, None, None
            else:
                con = database_connect()
                con.autocommit = True

                # Determine the id of the outputfile
                idnum_ = pd.read_sql(f'''SELECT idnum
                                        FROM outputfile
                                        WHERE filename='{fname}' ''', con)
                oid = idnum_.idnum[0]

                if self.quantity == 'radiance':
                    mech = ("mechanism = '" +
                            ", ".join(sorted([m for m in self.mechanism])) +
                            "'")
                    wave_ = sorted([w.value for w in self.wavelength])
                    wave = ("wavelength = '" +
                            ", ".join([str(w) for w in wave_]) +
                            "'")
                else:
                    mech = 'mechanism is NULL'
                    wave = 'wavelength is NULL'

                result = pd.read_sql(
                    f'''SELECT idnum, filename FROM uvvsmodels
                        WHERE out_idnum={oid} and
                              quantity = '{self.quantity}' and
                              orbit = {orb} and
                              dphi = {self.dphi.value} and
                              {mech} and
                              {wave}''', con)

                assert len(result) <= 1
                if len(result) == 1:
                    savefile = result.filename[0]
                    with open(savefile, 'rb') as f:
                        radiance, packets = pickle.load(f)
                    idnum = result.idnum[0]
                else:
                    radiance, packets, idnum = None, None, None

            return radiance, packets, idnum

    def create_model(self, data, outfile):
        # distance of s/c from planet
        dist_from_plan = np.sqrt(data.x**2 + data.y**2 + data.z**2)

        # Angle between look direction and planet.
        ang = np.arccos((-data.x*data.xbore - data.y*data.ybore -
                         data.z*data.zbore)/dist_from_plan)

        # Check to see if look direction intersects the planet anywhere
        asize_plan = np.arcsin(1./dist_from_plan)

        # Don't worry about lines of sight that don't hit the planet
        dist_from_plan[ang > asize_plan] = 1e30

        # Load the outputfile
        output = results_loadfile(outfile)
        radvel_sun = output.vy + output.vrplanet

        # Will base shadow on line of sight, not the packets
        out_of_shadow = np.ones_like(output.x)
        weight = results_packet_weighting(self, radvel_sun, output.frac,
                                          out_of_shadow, output.aplanet)

        xx_, yy_, zz_ = (np.zeros((2,len(data))), np.zeros((2,len(data))),
                         np.zeros((2,len(data))))
        xx_[1,:], yy_[1,:], zz_[1,:] = (data.xbore*10, data.ybore*10,
                                        data.zbore*10)
        xx = (data.x[np.newaxis,:] + xx_)
        yy = (data.y[np.newaxis,:] + yy_)
        zz = (data.z[np.newaxis,:] + zz_)

        xx_min = np.min(xx-0.5, axis=0)*self.unit
        yy_min = np.min(yy-0.5, axis=0)*self.unit
        zz_min = np.min(zz-0.5, axis=0)*self.unit
        xx_max = np.max(xx+0.5, axis=0)*self.unit
        yy_max = np.max(yy+0.5, axis=0)*self.unit
        zz_max = np.max(zz+0.5, axis=0)*self.unit

        radiance, packets = np.zeros(len(data)), np.zeros(len(data))

        for i,row in data.iterrows():
            j = i - min(data.index)
            # This removes the packets that aren't close to the los
            mask = ((output.x >= xx_min[j]) &
                    (output.x <= xx_max[j]) &
                    (output.y >= yy_min[j]) &
                    (output.y <= yy_max[j]) &
                    (output.z >= zz_min[j]) &
                    (output.z <= zz_max[j]))
            x_, y_, z_, w_, rvsun_ = (output.x[mask], output.y[mask],
                                      output.z[mask], weight[mask],
                                      radvel_sun[mask])

            # Distance from the spacecraft
            xpr = x_ - row.x*self.unit
            ypr = y_ - row.y*self.unit
            zpr = z_ - row.z*self.unit
            rpr = np.sqrt(xpr**2 + ypr**2 + zpr**2)

            # Packet-s/c boresight angle
            costheta = (xpr*row.xbore + ypr*row.ybore + zpr*row.zbore)/rpr
            costheta[costheta > 1] = 1.
            costheta[costheta < -1] = -1.

            inview = ((costheta >= np.cos(self.dphi)) &
                      (w_ > 0) &
                      (rpr < dist_from_plan[i]*self.unit))

            if np.any(inview):
                Apix = np.pi * (rpr[inview]*np.sin(self.dphi))**2
                wtemp = w_[inview]/Apix.to(u.cm**2)
                wtemp = wtemp.value
                if self.quantity == 'radiance':
                    # Determine if any packets are in shadow
                    # Projection of packet onto LOS
                    losrad = rpr[inview] * costheta[inview]

                    # Point along LOS the packet represents
                    xhit = row.x + row.xbore*losrad.value
                    yhit = row.y + row.ybore*losrad.value
                    zhit = row.z + row.zbore*losrad.value

                    rhohit = xhit**2 + zhit**2
                    out_of_shadow = (rhohit > 1) | (yhit < 0)
                    wtemp *= out_of_shadow

                    radiance[j] = np.sum(wtemp)
                    packets[j] = np.sum(inview)

        del output
        self.save(data, outfile, radiance, packets)

        return radiance, packets



import os, os.path
from .database_connect import database_connect
import sys
import types


def configfile():
    """Configure external resources used in the model

    Paths are saved in $HOME/.nexoclom
    * savepath = path where output files are saved
    * datapath = path where MESSENGER data is kept
    * database = name of the postgresql database to use
    """
    # Determne the savepath
    cfile = os.path.join(os.environ['HOME'], '.nexoclom')
    savepath = None
    if os.path.isfile(cfile):
        for line in open(cfile, 'r').readlines():
            key, value = line.split('=')
            if key.strip() == 'savepath':
                savepath = value.strip()
            else:
                pass
    else:
        pass

    if savepath is None:
        savepath = input('Where should outputs be saved: ')
        with open(cfile, 'a') as f:
            f.write(f'savepath = {savepath}\n')

        # Create save directory if necessary
        if not os.path.isdir(savepath):
            try:
                os.makedirs(savepath)
            except:
                assert 0, f'Could not create directory {savepath}'

    # Create the database if necessary
    database, port = database_connect(return_con=False)
    with database_connect(database='postgres') as con:
        cur = con.cursor()
        cur.execute('select datname from pg_database')
        dbs = [r[0] for r in cur.fetchall()]

        if database not in dbs:
            print(f'Creating database {database}')
            cur.execute(f'create database {database}')
        else:
            pass

    return savepath


def set_up_output_tables():
    con = database_connect()
    cur = con.cursor()

    # drop tables if necessary
    nextables = ['outputfile', 'geometry', 'sticking_info', 'forces',
                 'spatialdist', 'speeddist', 'angulardist', 'options',
                 'modelimages', 'uvvsmodels']
    cur.execute('select table_name from information_schema.tables')
    tables = [r[0] for r in cur.fetchall()]

    for tab in nextables:
        if tab in tables:
            cur.execute(f'''DROP table {tab}''')
        else:
            pass

    # create outputfile table
    cur.execute('''CREATE TABLE outputfile (
                       idnum SERIAL PRIMARY KEY,
                       filename text UNIQUE,
                       npackets bigint,
                       totalsource float,
                       creationtime timestamp NOT NULL)''')
    print('Created outputfile table')

    # create geometry table
    cur.execute('''CREATE TABLE geometry (
                       geo_idnum bigint PRIMARY KEY,
                       planet SSObject,
                       StartPoint SSObject,
                       objects SSObject ARRAY,
                       starttime timestamp,
                       phi real ARRAY,
                       subsolarpt point,
                       TAA float)''')
    print('Created geometry table')

    # Create sticking_info table
    cur.execute('''CREATE TABLE sticking_info (
                       st_idnum bigint PRIMARY KEY,
                       stickcoef float,
                       tsurf float,
                       stickfn text,
                       stick_mapfile text,
                       epsilon float,
                       n float,
                       tmin float,
                       emitfn text,
                       accom_mapfile text,
                       accom_factor float)''')
    print('Created sticking_info table')

    # create forces table
    cur.execute('''CREATE TABLE forces (
                       f_idnum bigint PRIMARY KEY,
                       gravity boolean,
                       radpres boolean)''')
    print('Created forces table')

    # create spatialdist table
    cur.execute('''CREATE TABLE spatialdist (
                       spat_idnum bigint PRIMARY KEY,
                       type text,
                       exobase float,
                       use_map boolean,
                       mapfile text,
                       longitude float[2],
                       latitude float[2])''')
    print('Created spatialdist table')

    # create table speeddist
    cur.execute('''CREATE TABLE speeddist (
                       spd_idnum bigint PRIMARY KEY,
                       type text,
                       vprob float,
                       sigma float,
                       U float,
                       alpha float,
                       beta float,
                       temperature float,
                       delv float)''')
    print('Created speeddist table')

    # create table angulardist
    cur.execute('''CREATE TABLE angulardist (
                       ang_idnum bigint PRIMARY KEY,
                       type text,
                       azimuth float[2],
                       altitude float[2],
                       n float)''')
    print('Created angulardist table')

    ## Skipping perturbvel and plasma_info for now

    # create table options
    cur.execute('''CREATE TABLE options (
                       opt_idnum bigint PRIMARY KEY,
                       endtime float,
                       resolution float,
                       at_once boolean,
                       atom text,
                       lifetime float,
                       fullsystem boolean,
                       outeredge float,
                       motion boolean,
                       streamlines boolean,
                       nsteps int)''')
    print('Created options table')

    # Create table for model images
    cur.execute('''CREATE TABLE modelimages (
                       idnum SERIAL PRIMARY KEY,
                       out_idnum bigint,
                       quantity text,
                       origin text,
                       dims float[2],
                       center float[2],
                       width float[2],
                       subobslongitude float,
                       subobslatitude float,
                       mechanism text,
                       wavelength text,
                       filename text)''')
    print('Created modelimages table')

    # # Create table for MESSENGER comparison
    cur.execute('''CREATE TABLE uvvsmodels (
                       idnum SERIAL PRIMARY KEY,
                       out_idnum bigint,
                       quantity text,
                       orbit int,
                       dphi float,
                       mechanism text,
                       wavelength text,
                       filename text)''')
    print('Created uvvsmodels table')
    con.close()


def configure_model(force=False):
    """Ensure the database and configuration file are set up for nexoclom."""
    if isinstance(sys.modules['psycopg2'], types.ModuleType):
        # Get database name and port
        database, port = database_connect(return_con=False)

        # Verify database is running
        status = os.popen('pg_ctl status').read()
        if 'no server running' in status:
            os.system(f'pg_ctl start -D $HOME/.postgres/main'
                      f'-l $HOME/.postgres/logfile -o "-p {port}"')
        else:
            pass

        # Determine if it is necessary to create the database tables
        nextables = ['outputfile', 'geometry', 'sticking_info', 'forces',
                     'spatialdist', 'speeddist', 'angulardist', 'options',
                     'modelimages', 'uvvsmodels']

        with database_connect() as con:
            cur = con.cursor()
            cur.execute('select table_name from information_schema.tables')
            tables = [r[0] for r in cur.fetchall()]
            there = [m in tables for m in nextables]

        if (False in there) or (force):
            set_up_output_tables()



import numpy as np
import astropy.units as u
from atomicdataMB import gValue
import mathMB


class ModelResult:
    def __init__(self, inputs):
        # Determine model scaling
        self.inputs = inputs
        self.filenames, self.packets, self.totalsource = inputs.findpackets()
        self.mod_rate = self.totalsource/inputs.options.endtime # pack/sec
        self.atoms_per_packet = (1e26/u.s)/self.mod_rate
        print(f'Total number of packets run = {self.packets}')
        print(f'Total source = {self.totalsource} packets')
        print(f'1 packet represents {self.atoms_per_packet} atoms')
        print(f'Model rate = {self.mod_rate} packets/sec')

def read_format(formatfile):
    format_ = {}
    with open(formatfile, 'r') as f:
        for line in f:
            try:
                p, v = line.split('=')
                format_[p.strip().lower()] = v.strip()
            except:
                pass
    return format_


def results_loadfile(filename):
    ''' Load the output and do some error checking

    It may be necessary at some point to add more functionality to this'''

    from .Output import Output

    # Load output
    output = Output.restore(filename)

    # Error checking
    try:
        output.x.unit
    except:
        assert 0, 'Output does not contain units like it should.'
    assert np.all(output.frac >= 0), 'Has f < 0'
    assert np.all(np.isfinite(output.frac)), 'Has non-finite f'

    # # Transform to moon-centric frame if necessary
    # if result.origin != result.inputs.geometry.planet:
    #     assert 0, 'Need to do transformation for a moon.'
    # else:
    #     origin = np.array([0., 0., 0.])*output.x.unit
    #     sc = 1.

    # Choose which packets to use
    # touse = output.frac >= 0 if keepall else output.frac > 0

    # packet positions relative to origin -- not rotated
    # pts_sun = np.array((output.x[touse]-origin[0],
    #                     output.y[touse]-origin[1],
    #                     output.z[touse]-origin[2]))*output.x.unit
    #
    # # Velocities relative to sun
    # vels_sun = np.array((output.vx[touse],
    #                      output.vy[touse],
    #                      output.vz[touse]))*output.vx.unit

    # Fractional content
    # frac = output.frac[touse]

    return output #, pts_sun, vels_sun, frac

def results_packet_weighting(result, radvel_sun, frac, out_of_shadow, aplanet):
    if result.quantity == 'column':
        weight = frac #* result.atoms_per_packet
    elif result.quantity == 'density':
        weight = frac #* result.atoms_per_packet
    elif result.quantity == 'radiance':
        if 'resscat' in result.mechanism:
            gg = np.zeros_like(frac)/u.s
            for w in result.wavelength:
                gval = gValue(result.inputs.options.atom, w, aplanet)
                gg += mathMB.interpu(radvel_sun, gval.velocity, gval.g)

            #weight_resscat = frac*result.atoms_per_packet*out_of_shadow*gg/1e6
            weight_resscat = frac*out_of_shadow*gg/1e6
        weight = weight_resscat # + other stuff

    assert np.all(np.isfinite(weight)), 'Non-finite weights'
    return weight



import os
import numpy as np
from astropy.time import Time
import astropy.units as u
from .Output import Output


def delete_files(filelist):
    ''' Delete output files and remove them from the database '''
    from .database_connect import database_connect

    with database_connect() as con:
        cur = con.cursor()

        for f in filelist:
            # Delete the file
            print(f)
            if os.path.exists(f):
                os.remove(f)

            # Remove from database
            cur.execute('''SELECT idnum FROM outputfile
                           WHERE filename = %s''', (f, ))
            idnum = cur.fetchone()[0]

            cur.execute('''DELETE FROM outputfile
                           WHERE idnum = %s''', (idnum, ))
            cur.execute('''DELETE FROM geometry
                           WHERE geo_idnum = %s''', (idnum, ))
            cur.execute('''DELETE FROM sticking_info
                           WHERE st_idnum = %s''', (idnum, ))
            cur.execute('''DELETE FROM forces
                           WHERE f_idnum = %s''', (idnum, ))
            cur.execute('''DELETE FROM spatialdist
                           WHERE spat_idnum = %s''', (idnum, ))
            cur.execute('''DELETE FROM speeddist
                           WHERE spd_idnum = %s''', (idnum, ))
            cur.execute('''DELETE FROM angulardist
                           WHERE ang_idnum = %s''', (idnum, ))
            cur.execute('''DELETE FROM options
                       WHERE opt_idnum = %s''', (idnum, ))
            print(f'Removed {idnum}: {os.path.basename(f)} from database')

            cur.execute('''SELECT idnum, filename FROM modelimages
                           WHERE out_idnum = %s''', (idnum, ))
            for mid, mfile in cur.fetchall():
                cur.execute('''DELETE from modelimages
                               WHERE idnum = %s''', (mid, ))
                if os.path.exists(mfile):
                    os.remove(mfile)

            cur.execute('''SELECT idnum, filename FROM uvvsmodels
                           WHERE out_idnum = %s''', (idnum, ))
            for mid, mfile in cur.fetchall():
                cur.execute('''DELETE from uvvsmodels
                               WHERE idnum = %s''', (mid, ))
                if os.path.exists(mfile):
                    os.remove(mfile)


def modeldriver(inputs, npackets, packs_per_it=None, overwrite=False,
                compress=True):
    '''
    Starting point for running the model
    INPUTS:
        inputs: Input object
        npackets: Total number of packets to run for this simulation
        overwrite: If True, delete existing files for this set of inputs
    '''

    t0_ = Time.now()
    print(f'Starting at {t0_}')

    if len(inputs.geometry.planet) != 1:
        print('Gravity and impact check not working for planets with moons.')
        sys.exit(1)

    # Determine how many packets have already been run
    outputfiles, totalpackets, _ = inputs.findpackets()
    print(f'Found {len(outputfiles)} files with {totalpackets} packets')

    if overwrite and (totalpackets > 0):
        # delete files and remove from database
        delete_files(outputfiles)
        totalpackets = 0
    else:
        pass

    npackets = int(npackets)
    ntodo = npackets - totalpackets
    if packs_per_it is None:
        packs_per_it = 100000 if inputs.options.streamlines else int(1e6)
        packs_per_it = min(ntodo, packs_per_it)
    else:
        packs_per_it = int(packs_per_it)

    if ntodo > 0:
        # Check to make sure at_once is set properly
        if inputs.options.streamlines and (inputs.options.at_once is False):
            raise RuntimeError

        # Determine how many iterations are needed
        nits = int(np.ceil(ntodo/packs_per_it))

        print('Running Model')
        print(f'Will compute {nits} iterations of {packs_per_it} packets.')

        # Do the iterations
        for _ in range(nits):
            tit0_ = Time.now()
            print(f'** Starting iteration #{_+1} of {nits} **')

            # Create an output object
            output = Output(inputs, packs_per_it, compress=compress)

            # Run the packets
            if inputs.options.streamlines:
                output.stream_driver()
            else:
                output.driver()

            # Save
            output.save()
            del output
            tit1_ = Time.now()
            print(f'** Completed iteration #{_+1} in {(tit1_-tit0_).sec} '
                  f'seconds.')
    else:
        pass

    t2_ = Time.now()
    dt_ = (t2_-t0_).sec
    if dt_ < 60:
        dt_ = f'{dt_} sec'
    elif dt_ < 3600:
        dt_ = f'{dt_/60} min'
    else:
        dt_ = f'{dt_/3600} hr'
    print(f'Model run completed in {dt_}')



import numpy as np
from .input_classes import AngularDist
from .source_distribution import angular_distribution

def bouncepackets(self, t1, x1, v1, f1, hhh):
    # This will need to be rewritten for satellite impacts

    # Determine where packets hit surface
    #srad = satrad[hhh]
    #r0 = tempR[hhh]
    x0 = x1[0,hhh]
    y0 = x1[1,hhh]
    z0 = x1[2,hhh]
    r0 = np.sqrt(x0**2 + y0**2 + z0**2)
    vx0 = v1[0,hhh]
    vy0 = v1[1,hhh]
    vz0 = v1[2,hhh]

    a = vx0**2 + vy0**2 + vz0**2  # = vv02
    b = 2*(x0*vx0 + y0*vy0 + z0*vz0)
    c = x0**2 + y0**2 + z0**2 - 1.
    dd = b**2 - 4*a*c
    assert np.all(dd >= 0)

    t0 = (-b - np.sqrt(b**2-4*a*c))/(2*a)
    t1 = (-b + np.sqrt(b**2-4*a*c))/(2*a)
    t = (t0 <= 0)*t0 + (t1 < 0)*t1

    # point where packet hit the surface
    x2 = x0 + vx0*t
    y2 = y0 + vy0*t
    z2 = z0 + vz0*t
    assert np.all(np.isfinite(x2))
    assert np.all(np.isfinite(y2))
    assert np.all(np.isfinite(z2))

    lonhit = (np.arctan2(x2, -y2) + 2*np.pi) % (2*np.pi)
    lathit = np.arcsin(z2)

    # put new coordinates into array
    x1[0,hhh] = x2
    x1[1,hhh] = y2
    x1[2,hhh] = z2

    # Determine rebound velocity
    #vv02 = vx0**2 + vy0**2 + vz0**2  # rplan/s
    PE = 2*self.GM*(1./r0 - 1)
    vv02 = a + PE
    vv02[vv02 < 0] = 0.
    assert np.all(np.isfinite(vv02))

    if self.inputs.sticking_info.emitfn.lower() == 'maxwellian':
        assert 0, 'Not set up yet'
        if self.inputs.sticking_info.Tsurf == 0:
            surftemp = SurfaceTemperatue(self.inputs.geometry,
                                         lonhit, lathit)
        else:
            pass # Need to set this up to
    elif self.inputs.sticking_info.emitfn.lower() == 'elastic scattering':
        vv2 = np.sqrt(vv02)
    else:
        assert 0, 'Emit function not set up yet'

    # Determine rebound angle

    angdist = AngularDist({'type':'costheta',
                           'altitude':f'0,{np.pi/2}',
                           'azimuth':f'0,{2*np.pi}'}, None)
#    angdist = AngularDist({'type':'radial'}, None)
    VV = angular_distribution(angdist, x1[:,hhh], vv2)

    # Rotate to proper position
    # to do

    # Put back into the arrays
    v1[:,hhh] = VV

    # Adjust the fractional values
    if self.inputs.sticking_info.stickcoef > 0:
        f1[hhh] *= 1 - self.inputs.sticking_info.stickcoef
    else:
        assert 0



import numpy as np
import psycopg2
import astropy.units as u
import astropy.units.astrophys as ua
from atomicdataMB import PhotoRate

class LossInfo:
    def __init__(self, atom, lifetime, aplanet):
        # Initialization
        self.photo = 0.
        self.eimp = 0.
        self.chX = 0.
        self.reactions = []

        if lifetime.value < 0:
            self.photo = np.abs(1./lifetime.value)
            self.reactions = 'Generic photo reaction.'
        else:
            photo = PhotoRate(atom, aplanet)
            self.photo = photo.rate.value
            self.reactions = photo.reactions['reaction'].values

            # Electron impact

            # Charge exchange

        if len(self.reactions) == 0:
            self.reactions = None

    def __len__(self):
        return len(self.reactions) if self.reactions is not None else 0

    def __str__(self):
        if len(self) == 0:
            print('No reactions included')
        elif len(self) == 1:
            print('Included Reaction: {}'.format(self.reactions[0]))
        else:
            print('\tIncluded Reactions: {}'.format(tuple(self.reactions)))
        if self.photo != 0:
            print('Photo Rate = {:0.2e} s'.
                  format(self.photo))
        if self.eimp != 0:
            print('Electron Impact Rate = {:0.2e} UNIT'.
                  format(self.eimp))
        if self.chX!= 0:
            print('Charge Exchange Rate = {:0.2e} UNIT'.
                  format(self.chX.value))

        return ''



"""Read the model inputs from a file and create the Input object.

The Input object is build from smaller objects defining different model
options.

Geometry
    Defines the Solar System geometry for the Input.

StickingInfo
    Defines the surface interactions.

Forces
    Set which forces act on model particles.

SpatialDist
    Define the initial spatial distribution of particles.

SpeedDist
    Define the initial speed distribution of particles.

AngularDist
    Define the initial angular distribtuion of particles.

Options
    Configure other model parameters
"""
import os
import os.path

import pandas as pd
from astropy.io import ascii

from .configure_model import configfile
from .database_connect import database_connect
from .input_classes import (Geometry, StickingInfo, Forces, SpatialDist,
                            SpeedDist, AngularDist, Options)
from .modeldriver import modeldriver
from .produce_image import ModelImage


class Input():
    def __init__(self, infile):
        """Read the input options from a file.

        **Parameters**
        infile
            Plain text file containing model input parameters.

        **Class Attributes**

        * geometry
        * sticking_info
        * forces
        * spatialdist
        * speeddist
        * angulardist
        * options

        **Class Methods**

        * findpackets()
        * run(npackets, overwrite=False, compress=True)
        * produce_image(format_, filenames=None)
        """
        # Read the configuration file
        self._savepath = configfile()

        # Read in the input file:
        self._inputfile = infile
        if os.path.isfile(infile):
            data = ascii.read(infile, delimiter='=', comment=';',
                              data_start=0, names=['Param', 'Value'])
        else:
            assert 0, 'Input file {} does not exist.'.format(infile)

        section = [d.split('.')[0].casefold() for d in data['Param']]
        param = [d.split('.')[1].casefold() for d in data['Param']]
        values = [v.split(';')[0].strip().casefold()
                  if ';' in v else v.casefold() for v in data['Value']]

        # Extract the geometry parameters
        gparam = {b:c for (a,b,c) in zip(section, param, values)
                  if a == 'geometry'}
        self.geometry = Geometry(gparam)

        # Extract the sticking information
        sparam = {b:c for a,b,c in zip(section, param, values)
                  if a == 'sticking_info'}
        self.sticking_info = StickingInfo(sparam)

        # Extract the forces
        fparam = {b:c for a,b,c in zip(section, param, values)
                  if a == 'forces'}
        self.forces = Forces(fparam)

        # Extract the spatial distribution
        sparam = {b:c for a,b,c in zip(section, param, values)
                  if a == 'spatialdist'}
        self.spatialdist = SpatialDist(sparam)

        # Extract the speed distribution
        sparam = {b:c for a,b,c in zip(section, param, values)
                  if a == 'speeddist'}
        self.speeddist = SpeedDist(sparam)

        # Extract the angular distribution
        aparam = {b:c for a,b,c in zip(section, param, values)
                  if a == 'angulardist'}
        self.angulardist = AngularDist(aparam, self.spatialdist)

        # Extract the options
        oparam = {b:c for a,b,c in zip(section, param, values)
                  if a == 'options'}
        self.options = Options(oparam, self.geometry.planet)
        
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        result = (self.geometry.__str__() + '\n' +
                  self.sticking_info.__str__() + '\n' +
                  self.forces.__str__() + '\n' +
                  self.spatialdist.__str__() + '\n' +
                  self.speeddist.__str__() + '\n' +
                  self.angulardist.__str__() + '\n' +
                  self.options.__str__())
        
        return result

    def findpackets(self):
        """ Search the database for identical inputs """
        georesult = self.geometry.search(startlist=None)
        if georesult is not None:
            stickresult = self.sticking_info.search(startlist=georesult)
        else:
            return [], 0, 0

        if stickresult is not None:
            forceresult = self.forces.search(startlist=stickresult)
        else:
            return [], 0, 0

        if forceresult is not None:
            spatresult = self.spatialdist.search(startlist=forceresult)
        else:
            return [], 0, 0

        if spatresult is not None:
            spdresult = self.speeddist.search(startlist=spatresult)
        else:
            return [], 0, 0

        if spdresult is not None:
            angresult = self.angulardist.search(startlist=spdresult)
        else:
            return [], 0, 0

        if angresult is not None:
            finalresult = self.options.search(startlist=angresult)
        else:
            return [], 0, 0

        if finalresult is not None:
            result_ = [str(s) for s in finalresult]
            resultstr = f"({', '.join(result_)})"
            with database_connect() as con:
                result = pd.read_sql(
                    f'''SELECT filename, npackets, totalsource
                        FROM outputfile
                        WHERE idnum in {resultstr}''', con)
            npackets = result.npackets.sum()
            totalsource = result.totalsource.sum()

            return result.filename.to_list(), npackets, totalsource
        else:
            return [], 0, 0

    def run(self, npackets, packs_per_it=None, overwrite=False, compress=True):
        modeldriver(self, npackets, packs_per_it=packs_per_it,
                    overwrite=overwrite, compress=compress)

    def produce_image(self, format_, filenames=None):
        return ModelImage(self, format_, filenames=filenames)



import os
import os.path
import numpy as np
import pickle
import astropy.units as u
from solarsystemMB import planet_dist
import mathMB
from .satellite_initial_positions import satellite_initial_positions
from atomicdataMB import RadPresConst
from .LossInfo import LossInfo
from .rk5 import rk5
from .bouncepackets import bouncepackets
from .source_distribution import (surface_distribution, speed_distribution,
                                  angular_distribution, surface_spot)
from .database_connect import database_connect


class Output:
    '''Keep track of packets initial and final positions
    x0, y0, z0, f0, vx0, vy0, vz0
    phi0, lat0, lon0
    time, x, y, z, vx, vy, vz
    index, npackets, totalsource
    '''

    def __init__(self, inputs, npackets, compress=True):
        self.inputs = inputs
        self.planet = inputs.geometry.planet

        # Keep track of whether output is compressed
        self.compress = compress

        # Determine spatial unit
        self.unit = u.def_unit('R_' + self.planet.object, self.planet.radius)

        # Change unit for GM
        self.GM = self.planet.GM.to(self.unit**3/u.s**2).value

        # Determine distance and radial velocity of planet relative to the Sun
        r, v_r = planet_dist(self.planet, self.inputs.geometry.taa)
        self.aplanet = r.value
        self.vrplanet = v_r.to(self.unit/u.s).value

        # Find the default reactions and datasets
        if inputs.options.lifetime.value <= 0:
            self.loss_info = LossInfo(inputs.options.atom,
                                      inputs.options.lifetime,
                                      self.aplanet)
        else:
            self.loss_info = None

        # Set up the radiation pressure
        if inputs.forces.radpres:
            radpres = RadPresConst(inputs.options.atom,
                                   self.aplanet)
            radpres.velocity = radpres.velocity.to(self.unit/u.s).value
            radpres.accel = radpres.accel.to(self.unit/u.s**2).value
            self.radpres = radpres
        else:
            self.radpres = None

        # Set up sticking
        # if inputs.sticking_info['stickcoef'] != 1:
        #     stick = sticking_setup(inputs)

        # Define the time that packets will run
        if inputs.options.at_once:
            time = np.ones(npackets) * inputs.options.endtime
        else:
            time = np.random.rand(npackets) * inputs.options.endtime

        self.time0 = time.copy()
        self.time = time.copy()

        # Define the fractional content
        self.f0 = np.ones(npackets)
        self.frac = np.ones(npackets)

        self.npackets = npackets
        self.totalsource = np.sum(self.f0)
        self.LossFrac = np.zeros(npackets)

        # Determine initial satellite positions if necessary
        if self.planet.moons is not None:
            sat_init_pos = satellite_initial_positions(inputs)
        else:
            pass

        # Determine initial source distribution
        self.source_distribution()

    def __str__(self):
        print('Contents of output:')
        print('\tPlanet = {}'.format(self.planet.object))
        print('\ta_planet = {}'.format(self.aplanet))
        print('\tvr_planet = {}'.format(self.vrplanet))
#        print(self.loss_info)
#        print('\tNumber of Packets: {}'.format(self.npackets))
#        print('\tUnits of time: {}'.format(self.time.unit))
#        print('\tUnits of distance: {}'.format(self.X0.unit))
#        print('\tUnits of velocity: {}'.format(self.V0.unit))
        return ''

    def __len__(self):
        return self.npackets

    def __getitem__(self, keys):
        self.time = self.time[keys]
        self.frac = self.frac[keys]
#        self.LossFrac = self.LossFrac[keys]
        self.x = self.x[keys]
        self.y = self.y[keys]
        self.z = self.z[keys]
        self.vx = self.vx[keys]
        self.vy = self.vy[keys]
        self.vz = self.vz[keys]
        if 'index' in self.__dict__.keys():
            self.index = self.index[keys]

    def source_distribution(self):
        # Determine spatial distribution
        if self.inputs.spatialdist.type == 'surface':
            X0, lon, lat = surface_distribution(self.inputs,
                                                self.npackets,
                                                self.unit)
        elif self.inputs.spatialdist.type == 'surfacespot':
            X0, lon, lat = surface_spot(self.inputs,
                                        self.npackets,
                                        self.unit)
        else:
            assert 0, 'Spatial Distribution {} not supported.'.format(
                self.inputs.spatialdist.type)

        # Choose a speed for each packet
        v0 = speed_distribution(self.inputs, self.npackets)
        if v0 is not None:
            v0 = v0.to(self.unit/u.s)

            # Choose direction for each packet
            V0 = angular_distribution(self.inputs, X0.value, v0.value)
            if V0 is not None:
                V0 *= self.unit/u.s
            else:
                pass
        else:
            V0 = None

        # Perturbation function
        # Not installed yet

        # Rotate everything to proper position for running the model
        if (self.inputs.geometry.planet.object !=
                self.inputs.geometry.startpoint):
            assert 0, 'Not set up yet'
        else:
            pass

        self.x0 = X0[0,:]
        self.y0 = X0[1,:]
        self.z0 = X0[2,:]
        self.vx0 = V0[0,:]
        self.vy0 = V0[1,:]
        self.vz0 = V0[2,:]

        self.x = X0[0,:]
        self.y = X0[1,:]
        self.z = X0[2,:]
        self.vx = V0[0,:]
        self.vy = V0[1,:]
        self.vz = V0[2,:]

    def driver(self):
        assert 0, 'Needs to be verified.'
        # Set up the step sizes
        hall = np.zeros(self.npackets) + 1000.

        count = 0  # Number of steps taken
        eps = 1e-10  # impact check delta

        # These control how quickly the stepsize is increased or decreased
        # between iterations
        safety = 0.95
        shrink = -0.25
        grow = -0.2

        # yscale = scaling parameter for each variable
        #     x,y,z ~ R_plan
        #     vx, vy, vz ~ 1 km/s (1/R_plan R_plan/s)
        #     frac ~ exp(-t/lifetime) ~ mean(frac)
        rest = self.inputs.options.resolution
        resx = self.inputs.options.resolution
        resv = 0.1*self.inputs.options.resolution
        resf = self.inputs.options.resolution

        #########################################################
        # Keep taking RK steps until every packet has reached the
        # time of "image taken"
        #########################################################

        T = self.time.value
        X = np.array([self.x, self.y, self.z])
        V = np.array([self.vx, self.vy, self.vz])
        Frac = self.frac
        LossFrac = self.LossFrac

        numb = []
        moretogo = (T > rest) & (Frac > 0.)
        while moretogo.any():
            # Save old values
            # This is used for determining if anything hits the rings
            oldT = T[moretogo].copy()
            oldX = X[:, moretogo].copy()
            oldV = V[:, moretogo].copy()
            oldFrac = Frac[moretogo].copy()

            T0 = T[moretogo]
            X0 = X[:, moretogo]
            V0 = V[:, moretogo]
            F0 = Frac[moretogo]
            Loss0 = LossFrac[moretogo]

            h = hall[moretogo]
            assert np.all(h >= 0), '\n\tNegative values of h'

            # Adjust stepsize to be no more than time remaining
            h = (T0 >= h)*h + (T0 < h)*T0

            # Run the rk5 step
            t1, x1, v1, f1, delx, delv, delf = rk5(T0, X0, V0, F0, h, self)

            # Do the error check
            # scale = a_tol + |y|*r_tol
            #   for x: a_tol = r_tol = resolution
            #   for v: a_tol = r_tol = resolution/10.-require v more precise
            #   for f: a_tol = 0.01, r_tol = 0 -> frac tol = 1%
            #      -> f not implemented this way
            scalex = resx + np.abs(x1)*resx
            scalev = resv + np.abs(v1)*resv
            scalef = resf + np.abs(f1)*resf

            # Difference relative to acceptable difference
            delx /= scalex
            delv /= scalev
            delf /= scalef

            # Maximum error for each packet
            xerrmax = delx.max(axis=0)
            verrmax = delv.max(axis=0)
            errmax = np.maximum(xerrmax, verrmax)
            errmax = np.maximum(errmax, delf)

            # error check
            assert np.all(np.isfinite(errmax)), '\n\tInfinite values of emax'

            # Make sure no negative frac
            m = (f1 < 0) & (errmax < 1)
            assert not np.any(m), 'Found new values of frac that are negative'

            # Make sure frac doesn't increase
            m = (f1-oldFrac > scalef) & (errmax > 1)
            if np.any(m):
                errmax[m] = 1.1

            # Check where difference is very small. Adjust step size
            noerr = errmax < 1e-7
            errmax[noerr] = 1
            h[noerr] *= 10

            # Put the post-step values in
            g = errmax < 1.0
            b = (errmax >= 1.0)
            numb.append(np.sum(b))

            if np.any(g):
                t_ = t1[g]
                x_ = x1[:, g]
                v_ = v1[:, g]
                f_ = f1[g]
                h_ = safety*h[g]*errmax[g]**grow

                # Impact Check
                tempR = np.linalg.norm(x_, axis=0)
                # satrad = np.ones_like(tempR)
                hhh = (tempR.value - 1) < 0
                if np.any(hhh):
                    if self.inputs.sticking_info.stickcoef == 1.:
                        f_[hhh] = 0.
                    else:
                        bouncepackets(self, t_, x_, v_, f_, hhh)

                # Check for escape
                if not self.inputs.options.fullsystem:
                    f_[tempR > self.inputs.options.outeredge] = 0

                # Check for vanishing
                f_[f_ < 1e-10] = 0

                # set remaining time = 0 for packets that are done
                t_[f_ == 0] = 0.

                # Put new values into arrays
                T0[g] = t_
                X0[:, g] = x_
                V0[:, g] = v_
                F0[g] = f_
                h[g] = h_

            if np.any(b):
                # Don't adjust the bad value, but do fix the stepsize
                htemp = safety*h[b]*errmax[b]**shrink
                assert np.all(np.isfinite(htemp)), '\n\tInfinite values of h'

                # Don't let step size drop below 1/10th previous step size
                h[b] = np.maximum(htemp, 0.1*h[b])

            assert np.all(h >= 0), '\n\tNegative values of h'

            # Insert back into the original arrays
            T[moretogo] = T0
            X[:, moretogo] = X0
            V[:, moretogo] = V0
            Frac[moretogo] = F0
            LossFrac[moretogo] = Loss0
            hall[moretogo] = h

            # Find which packets still need to run
            moretogo = (T > rest) & (Frac > 0.)
            if count % 100 == 0:
                print(f'Step {count}. {np.sum(moretogo)} more to go\n'
                      f'\th: {mathMB.minmaxmean(h)}')
            count += 1

        # Put everything back into output
        self.time = T*u.s
        self.x = X[0, :]*self.unit
        self.y = X[1, :]*self.unit
        self.z = X[2, :]*self.unit
        self.vx = V[0, :]*self.unit/u.s
        self.vy = V[1, :]*self.unit/u.s
        self.vz = V[2, :]*self.unit/u.s
        self.frac = Frac
        # self.LossFrac = lf0

        # Add units back in
        self.aplanet *= u.au
        self.vrplanet *= self.unit/u.s
        self.vrplanet = self.vrplanet.to(u.km/u.s)
        self.GM *= self.unit**3/u.s**2

    def stream_driver(self):
        T = self.time.value
        X = np.array([self.x, self.y, self.z])
        V = np.array([self.vx, self.vy, self.vz])

        Frac = self.frac
        LossFrac = self.LossFrac
        npack = len(T)

        # Arrays to store the outputs
        xx0 = np.zeros((3, npack, self.inputs.options.nsteps))
        vv0 = np.zeros((3, npack, self.inputs.options.nsteps))
        ff0 = np.zeros((npack, self.inputs.options.nsteps))
        tt0 = np.zeros((npack, self.inputs.options.nsteps))

        # Add initial values to the arrays
        tt0[:,0] = T
        xx0[:,:,0] = X
        vv0[:,:,0] = V
        ff0[:,0] = Frac

        #  step size and counters
        dt = self.inputs.options.endtime.value/(self.inputs.options.nsteps-1)
        h = dt
        curtime = self.inputs.options.endtime.value
        ct = 1

        eps = 1e-10

        moretogo = Frac > 0
        while (curtime > 0) and (moretogo.any()):
            t0 = T[moretogo]
            x0 = X[:, moretogo]
            v0 = V[:, moretogo]
            f0 = Frac[moretogo]
            oldF = f0.copy()

            # Error checks
            assert (f0 != 0).all(), 'Should not have 0 frac'
            assert np.isfinite(x0).all(), 'Infinite X values'
            assert np.isfinite(v0).all(), 'Infinite V values'

            # Run the rk5 step
            t1, x1, v1, f1, _, _, _ = rk5(t0, x0, v0, f0, h, self)

            # Check for surface impacts
            tempR = np.linalg.norm(x1, axis=0)
            # satrad = np.ones_like(tempR)
            hit_surf = (tempR - 1.) < 0

            if np.any(hit_surf):
                if self.inputs.sticking_info.stickcoef == 1:
                    f1[hit_surf] = 0.
                else:
                    bouncepackets(self, t1, x1, v1, f1, hit_surf)

            # Check for escape
            if not self.inputs.options.fullsystem:
                f1[tempR > self.inputs.options.outeredge] = 0

            # Check to see if any packets have shrunk out of existence
            f1[f1 < 1e-10] = 0.

            # Set remaining time = 0 for packets that are done
            t1[f1 == 0] = 0.

            # Put new values back into the original array
            T[moretogo] = t1
            X[:,moretogo] = x1
            V[:,moretogo] = v1
            Frac[moretogo] = f1

            # Save the results of this step
            tt0[:,ct] = T
            xx0[:,:,ct] = X
            vv0[:,:,ct] = V
            ff0[:,ct] = Frac

            # Check to see what still needs to be done
            moretogo = Frac > 0
            done = moretogo.any() is False

            if (ct % 100) == 0:
                print(ct, curtime, np.sum(moretogo))

            # Update the times
            ct += 1
            curtime -= dt

        # Put everything back into output
        allpacks = npack*self.inputs.options.nsteps
        self.time = tt0.reshape(allpacks)*u.s
        self.x = xx0[0, :, :].reshape(allpacks)*self.unit
        self.y = xx0[1, :, :].reshape(allpacks)*self.unit
        self.z = xx0[2, :, :].reshape(allpacks)*self.unit
        self.vx = vv0[0, :, :].reshape(allpacks)*self.unit/u.s
        self.vy = vv0[1, :, :].reshape(allpacks)*self.unit/u.s
        self.vz = vv0[2, :, :].reshape(allpacks)*self.unit/u.s
        self.frac = ff0.reshape(allpacks)
        # self.LossFrac = lf0

        self.totalsource *= self.inputs.options.nsteps
        index = np.mgrid[0:npack, 0:self.inputs.options.nsteps]
        self.index = index[0,:,:].reshape(allpacks)

        # Add units back in
        self.aplanet *= u.au
        self.vrplanet *= self.unit/u.s
        self.vrplanet = self.vrplanet.to(u.km/u.s)
        self.GM *= self.unit**3/u.s**2

    def determine_filename(self):
        '''Come up with a filename for the model'''

        # TAA for observation
        if self.planet.object == 'Mercury':
            taastr = '{:03.0f}'.format(
                      np.round(self.inputs.geometry.taa.to(u.deg).value))
        else:
            assert 0, 'Filename not set up for anything but Mercury'

        # Come up with a path name
        pathname = os.path.join(self.inputs._savepath,
                                self.planet.object,
                                self.inputs.options.atom,
                                self.inputs.spatialdist.type,
                                self.inputs.speeddist.type,
                                taastr)

        # Make the path if necessary
        if os.path.exists(pathname) == 0:
            os.makedirs(pathname)
        numstr = '{:010d}'.format(self.idnum)
        filename = f'{numstr}.pkl'

        return pathname, filename

    def save(self):
        # Add output into database
        con = database_connect()
        cur = con.cursor()

        # Save filename and determine idnum
        cur.execute('''INSERT into outputfile (npackets,
                                               totalsource,
                                               creationtime)
                       VALUES (%s, %s, NOW())''', (self.npackets,
                                                   self.totalsource))
        cur.execute('''SELECT idnum
                       FROM outputfile
                       WHERE filename is Null''')
        assert cur.rowcount == 1, 'Outputs with NULL filename'
        self.idnum = cur.fetchone()[0]

        # Get a filename for the new savefile
        savefile = os.path.join(*self.determine_filename())
        print('Saving output as {}'.format(savefile))

        # save the geometry
        geometry = self.inputs.geometry
        objs = [obj.object for obj in geometry.objects]
        objs.sort()

        phi = [p.value for p in geometry.phi]
        cur.execute('''INSERT into geometry
                       VALUES (%s, %s, %s, %s::SSObject[],
                               %s, %s, '(%s, %s)', %s)''',
                    (self.idnum,
                     geometry.planet.object,
                     geometry.startpoint,
                     objs,
                     geometry.time,
                     phi,
                     geometry.subsolarpoint[0].value,
                     geometry.subsolarpoint[1].value,
                     geometry.taa.value))

        # Save the sticking_info
        sticking_info = self.inputs.sticking_info
        cur.execute('''INSERT into sticking_info
                       VALUES (%s, %s, %s, %s, %s, %s,
                               %s, %s, %s, %s, %s)''',
                    (self.idnum,
                     sticking_info.stickcoef,
                     sticking_info.tsurf,
                     sticking_info.stickfn,
                     sticking_info.stick_mapfile,
                     sticking_info.epsilon,
                     sticking_info.n,
                     sticking_info.tmin,
                     sticking_info.emitfn,
                     sticking_info.accom_mapfile,
                     sticking_info.accom_factor))

        # Save the forces
        forces = self.inputs.forces
        cur.execute('''INSERT into forces
                       VALUES (%s, %s, %s)''',
                    (self.idnum, forces.gravity, forces.radpres))

        # Save the spatial distribution
        spatdist = self.inputs.spatialdist
        if spatdist.longitude is None:
            lon = 0., 0.
        else:
            lon = spatdist.longitude[0].value, spatdist.longitude[1].value
        if spatdist.latitude is None:
            lat = 0., 0.
        else:
            lat = spatdist.latitude[0].value, spatdist.latitude[1].value

        cur.execute('''INSERT into spatialdist
                       VALUES (%s, %s, %s, %s, %s,
                               ARRAY[%s,%s], ARRAY[%s,%s])''',
                    (self.idnum,
                     spatdist.type,
                     spatdist.exobase,
                     spatdist.use_map,
                     spatdist.mapfile,
                     lon[0], lon[1],
                     lat[0], lat[1]))

        # Save the speed distribution
        speeddist = self.inputs.speeddist
        vprob = None if speeddist.vprob is None else speeddist.vprob.value
        sigma = None if speeddist.sigma is None else speeddist.sigma.value
        U = None if speeddist.U is None else speeddist.U.value
        temperature = (None if speeddist.temperature is None
                       else speeddist.temperature.value)
        delv = None if speeddist.delv is None else speeddist.delv.value

        cur.execute('''INSERT into speeddist
                       VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s)''',
                    (self.idnum,
                     speeddist.type,
                     vprob,
                     sigma,
                     U,
                     speeddist.alpha,
                     speeddist.beta,
                     temperature,
                     delv))

        # save the angular distribution
        angdist = self.inputs.angulardist
        if angdist.azimuth is None:
            az = 'NULL'
        else:
            az0 = angdist.azimuth[0].value
            az1 = angdist.azimuth[1].value
            az = f'ARRAY[{az0}, {az1}]'

        if angdist.altitude is None:
            alt = 'NULL'
        else:
            alt0 = angdist.altitude[0].value
            alt1 = angdist.altitude[1].value
            alt = f'ARRAY[{alt0}, {alt1}]'

        n = None if angdist.n is None else angdist.n

        cur.execute(f'''INSERT into angulardist
                        VALUES (%s, %s, {az}, {alt}, %s)''',
                    (self.idnum,
                     angdist.type,
                     n))

        # Save the options
        options = self.inputs.options
        cur.execute('''INSERT into options
                       VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)''',
                    (self.idnum,
                     options.endtime.value,
                     options.resolution,
                     options.at_once,
                     options.atom,
                     options.lifetime.value,
                     options.fullsystem,
                     options.outeredge,
                     options.motion,
                     options.streamlines,
                     options.nsteps))

        cur.execute('''UPDATE outputfile
                       SET filename=%s
                       WHERE idnum=%s''', (savefile, self.idnum))
        con.close()

        # Remove frac = 0
        if self.compress:
            self[self.frac > 0]

        # Put the data into a better form for saving
        self.time0 = self.time0.value.astype(np.float32)
        self.f0 = self.f0.astype(np.float32)
        self.x0 = self.x0.value.astype(np.float32)
        self.y0 = self.y0.value.astype(np.float32)
        self.z0 = self.z0.value.astype(np.float32)
        self.vx0 = self.vx0.value.astype(np.float32)
        self.vy0 = self.vy0.value.astype(np.float32)
        self.vz0 = self.vz0.value.astype(np.float32)

        self.time = self.time.value.astype(np.float32)
        self.frac = self.frac.astype(np.float32)
        self.x = self.x.value.astype(np.float32)
        self.y = self.y.value.astype(np.float32)
        self.z = self.z.value.astype(np.float32)
        self.vx = self.vx.value.astype(np.float32)
        self.vy = self.vy.value.astype(np.float32)
        self.vz = self.vz.value.astype(np.float32)
        self.LossFrac = self.LossFrac.astype(np.float32)

        # Save output as a pickle
        with open(savefile, 'wb') as f:
            pickle.dump(self, f, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def restore(cls, filename, reform=False):
        output = pickle.load(open(filename, 'rb'))
        output.time0 *= u.s
        output.time *= u.s
        output.x0 *= output.unit
        output.y0 *= output.unit
        output.z0 *= output.unit
        output.vx0 *= output.unit/u.s
        output.vy0 *= output.unit/u.s
        output.vz0 *= output.unit/u.s
        output.x *= output.unit
        output.y *= output.unit
        output.z *= output.unit
        output.vx *= output.unit/u.s
        output.vy *= output.unit/u.s
        output.vz *= output.unit/u.s

        if reform:
            index_ = list(set(output.index))
            n, m = len(index_), output.inputs.options.nsteps
            x = np.ndarray((n,m))
            y = np.ndarray((n,m))
            z = np.ndarray((n,m))
            vx = np.ndarray((n,m))
            vy = np.ndarray((n,m))
            vz = np.ndarray((n,m))
            frac, index = np.ndarray((n,m)), np.ndarray((n,m))
            time = np.ndarray((n,m))
            for i, ind in enumerate(index_):
                q = output.index == ind
                x[i,:sum(q)] = output.x[q]
                y[i,:sum(q)] = output.y[q]
                z[i,:sum(q)] = output.z[q]
                vx[i,:sum(q)] = output.vx[q]
                vy[i,:sum(q)] = output.vy[q]
                vz[i,:sum(q)] = output.vz[q]
                frac[i,:sum(q)] = output.frac[q]
                index[i,:sum(q)] = output.index[q]
                time[i,:sum(q)] = output.time[q]

            output.x = x
            output.y = y
            output.z = z
            output.vx = vx
            output.vy = vy
            output.vz = vz
            output.frac = frac
            output.index = index
            output.time = time

        return output



import os, os.path
import numpy as np
import pickle
import astropy.units as u
import astropy.constants as const
import mathMB
from atomicdataMB import atomicmass


def sputdist(velocity, U, alpha, beta, atom):
    matom = atomicmass(atom)
    v_b = np.sqrt(2*U/matom)
    v_b = v_b.to(u.km/u.s)
    f_v = velocity**(2*beta+1) / (velocity**2 + v_b**2)**alpha
    f_v /= np.max(f_v)
    return f_v.value


def MaxwellianDist(velocity, temperature, atom):
    vth2 = 2*temperature*const.k_B/atomicmass(atom)
    vth2 = vth2.to(u.km**2/u.s**2)
    f_v = velocity**3 * np.exp(-velocity**2/vth2)
    f_v /= np.max(f_v)
    return f_v.value


def xyz_from_lonlat(lon, lat, isplan, unit, exobase):
    if isplan:
        # Starting at a planet
        # 0 deg longitude = subsolar pt. = (0, -1, 0)
        # 90 deg longitude = dusk pt. = (1, 0, 0)
        # 270 deg longitude = dawn pt. = (-1, 0, 0)
        x0 = exobase * np.sin(lon) * np.cos(lat)
        y0 = -exobase * np.cos(lon) * np.cos(lat)
        z0 = exobase * np.sin(lat)
    else:
        # Starting at a satellite
        # 0 deg longitude = sub-planet pt. = (0, -1, 0)
        # 90 deg longitude = leading pt. = (-1, 0, 0)
        # 270 deg longitude = trailing pt. = (1, 0, 0)
        x0 = -exobase * np.sin(lon) * np.cos(lat)
        y0 = -exobase * np.cos(lon) * np.cos(lat)
        z0 = exobase * np.sin(lat)

    X0 = np.array([x0, y0, z0])*unit

    # Error checking
    assert np.all(np.isfinite(X0)), 'Non-Finite values of X0'
    return X0


def surface_distribution(inputs, npack, unit):
    '''
    Distribution packets on a sphere with radius r = SpatialDist.exobase

    Returns (x0, y0, z0, lon0, lat0)
    for satellites, assumes satellite is at phi=0

    Testing notes:
    1) Basic function with lon = [0,360], lat = [-90,90]
    2) Spatial map given
    3) Longitude, latitude range
    3) Longitude, latitude range, lon1 > lon0
    '''

    SpatialDist = inputs.spatialdist
    if SpatialDist.use_map:
        # Choose lon, lat based on predetermined map
        if SpatialDist.mapfile.endswith('.pkl'):
            with open(SpatialDist.mapfile, 'rb') as mapfile:
                sourcemap = pickle.load(mapfile)
        elif SpatialDist.mapfile.endswith('.sav'):
            from scipy.io import readsav
            sourcemap_ = readsav(SpatialDist.mapfile)['sourcemap']
            sourcemap = {'longitude':sourcemap_['longitude'][0]*u.rad,
                         'latitude':sourcemap_['latitude'][0]*u.rad,
                         'map':sourcemap_['map'][0].transpose()}
        else:
            assert 0, 'Mapfile is the wrong format.'

        lon, lat = mathMB.random_deviates_2d(sourcemap['map'],
                                           sourcemap['longitude'],
                                           np.sin(sourcemap['latitude']),
                                           npack)
        lat = np.arcsin(lat)
    else:
        # Choose the latitude: f(lat) = cos(lat)
        lat0 = SpatialDist.latitude
        if lat0[0] == lat0[1]:
            lat = np.zeros(npack)+lat0[0]
        else:
            ll = (np.sin(lat0[0]), np.sin(lat0[1]))
            sinlat = ll[0] + (ll[1]-ll[0]) * np.random.rand(npack)
            lat = np.arcsin(sinlat)

        # Choose the longitude: f(lon) = 1/(lonmax-lonmin)
        lon0 = SpatialDist.longitude
        if lon0[0] > lon0[1]:
            lon0 = [lon0[0], lon0[1]+2*np.pi*u.rad]
        lon = ((lon0[0] + (lon0[1]-lon0[0]) * np.random.rand(npack)) %
                (2*np.pi*u.rad))

    X0 = xyz_from_lonlat(lon, lat, inputs.geometry.planet.type == 'Planet',
                         unit, SpatialDist.exobase)
    return X0, lon, lat

def surface_spot(inputs, npackets, unit):
    '''Create a spot that drops off exponentially from (lon0, lat0)'''

    lon0, sigma = inputs.spatialdist.longitude
    lat0, _ = inputs.spatialdist.latitude

    spot0 = (np.sin(lon0)*np.cos(lat0),
             -np.cos(lon0)*np.cos(lat0),
             np.sin(lat0))
    longitude = np.linspace(0, 2*np.pi, 361)*u.rad
    latitude = np.linspace(-np.pi/2, np.pi/2, 181)*u.rad

    ptsx = np.outer(np.sin(longitude), np.cos(latitude))
    ptsy = -np.outer(np.cos(longitude), np.cos(latitude))
    ptsz = -np.outer(np.ones_like(longitude), np.sin(latitude))

    cosphi = ptsx*spot0[0] + ptsy*spot0[1] + ptsz*spot0[2]
    cosphi[cosphi > 1] = 1
    cosphi[cosphi < -1] = -1
    phi = np.arccos(cosphi)
    sourcemap = np.exp(-phi/sigma)

    lon, lat = mathMB.random_deviates_2d(sourcemap, longitude,
                                         np.sin(latitude), npackets)
    lat = np.arcsin(lat)

    X0 = xyz_from_lonlat(lon, lat, inputs.geometry.planet.type == 'Planet',
                         unit, inputs.spatialdist.exobase)

    return X0, lon, lat


def idlversion(inputs, unit):
    from scipy.io import readsav

    # Determine IDL files
    path, _ = os.path.split(__file__)
    rfile = os.path.join(path, 'data', 'modeloutput_search_routines.sav')
    vfile = os.path.join(path, 'data', 'modeloutput_search_variables.sav')

    with open('make_file_list.pro', 'w') as f:
        #profname = f.name
        #tempdir, _ = os.path.split(f.name)
        #filelistfile = os.path.join(tempdir, 'filelist.dat')
        filelistfile = 'filelist.dat'
        f.write(f'''
pro make_file_list

restore, '{vfile}'
defsysv, '!model', modelvar
filelist = modeloutput_search('{inputs.spatialdist.mapfile}')
openw, 1, '{filelistfile}'
for i=0,n_elements(filelist)-1 do printf, 1, filelist[i]
close, 1

end''')

    # Search for the IDL files
    cmd = (f'/Applications/exelis/idl/bin/idl '
           f'''-e "restore, '{rfile}' & make_file_list" ''')
    os.system(cmd)

    # Load the initial values
    x0 = None
    idloutputfiles = open('filelist.dat', 'r').readlines()
    print(f'{len(idloutputfiles)} IDL sav files found')
    for savfile in idloutputfiles:
        idl = readsav(savfile.strip())
        idlout = idl['output']

        index = sorted(list(set(idlout['index'][0])))
        x_ = [idlout['x0'][0][idlout['index'][0] == i][0] for i in index]
        y_ = [idlout['y0'][0][idlout['index'][0] == i][0] for i in index]
        z_ = [idlout['z0'][0][idlout['index'][0] == i][0] for i in index]
        vx_ = [idlout['vx0'][0][idlout['index'][0] == i][0] for i in index]
        vy_ = [idlout['vy0'][0][idlout['index'][0] == i][0] for i in index]
        vz_ = [idlout['vz0'][0][idlout['index'][0] == i][0] for i in index]
#        f_ = [idlout['f0'][0][idlout['index'][0] == i][0] for i in index]
        f_ = np.ones_like(x_)

        if x0 is None:
            x0 = np.array(x_)
            y0 = np.array(y_)
            z0 = np.array(z_)
            vx0 = np.array(vx_)
            vy0 = np.array(vy_)
            vz0 = np.array(vz_)
            f = np.array(f_)
        else:
            x0 = np.append(x0, np.array(x_))
            y0 = np.append(y0, np.array(y_))
            z0 = np.append(z0, np.array(z_))
            vx0 = np.append(vx0, np.array(vx_))
            vy0 = np.append(vy0, np.array(vy_))
            vz0 = np.append(vz0, np.array(vz_))
            f = np.append(f, np.array(f_))

    X = [x0, y0, z0]*unit
    V = [vx0, vy0, vz0]*unit/u.s
    F = f

    return X, V, F

def speed_distribution(inputs, npackets):
    SpeedDist = inputs.speeddist

    if inputs.spatialdist.type == 'idlversion':
        return None
    else:
        pass

    if SpeedDist.type.lower() == 'gaussian':
        if SpeedDist.sigma == 0.:
            v0 = np.zeros(npackets)*u.km/u.s + SpeedDist.vprob
        else:
            v0 = np.random.normal(SpeedDist.vprob.value,
                                  SpeedDist.sigma.value,
                                  npackets)
            v0 *= SpeedDist.vprob.unit
    elif SpeedDist.type == 'sputtering':
        velocity = np.linspace(.1, 50, 5000)*u.km/u.s
        f_v = sputdist(velocity, SpeedDist.U, SpeedDist.alpha,
                       SpeedDist.beta, inputs.options.atom)

        v0 = (mathMB.random_deviates_1d(velocity, f_v.value, npackets) *
              velocity.unit)
    elif SpeedDist.type == 'maxwellian':
        if SpeedDist.temperature != 0:
            # Use a constant temperature
            amass = atomicmass(inputs.options.atom)
            v_th = np.sqrt(2*SpeedDist.temperature*const.k_B/amass)
            v_th = v_th.to(u.km/u.s)
            velocity = np.linspace(0.1*u.km/u.s, v_th*5, 5000)
            f_v = MaxwellianDist(velocity, SpeedDist.temperature,
                                 inputs.options.atom)
            v0 = (mathMB.random_deviates_1d(velocity.value, f_v, npackets) *
                  velocity.unit)
        else:
            # Use a surface temperature map
            # Need to write this
            assert 0, 'Not implemented yet'
    elif SpeedDist.type == 'flat':
        v0 = (np.random.rand(npackets)*2*SpeedDist.delv +
               SpeedDist.vprob - SpeedDist.delv)
    else:
        # Need to add more distributions
        assert 0, 'Distribtuion does not exist'

    assert np.all(np.isfinite(v0)), 'Infinite values for v0'

    return v0


def angular_distribution(inputs, X0, vv):
    if inputs.spatialdist.type == 'idlversion':
        return None
    else:
        pass

    npackets = len(vv)

    AngularDist = inputs.angulardist

    if AngularDist.type == 'none':
        pass
    elif AngularDist.type == 'radial':
        alt = np.zeros(npackets) + np.pi/2. # All packets going directly up
        az = np.zeros(npackets)
    elif AngularDist.type == 'isotropic':
        # Choose the altitude -- f(alt) = cos(alt)
        alt0 = AngularDist.altitude
        aa = (np.sin(alt0[0]), np.sin(alt0[1]))
        sinalt = np.random.rand(npackets) * (aa[1] - aa[0]) + aa[0]
        alt = np.arcsin(sinalt)

        # Choose the azimuth -- f(az) = 1/(azmax-azmin)
        az0, az1 = AngularDist.azimuth
        m = (az0, az1) if az0 < az1 else (az1, az0+2*np.pi)
        az = (m[0] + (m[1]-m[0])*np.random.rand(npackets)) % (2*np.pi*u.rad)
    elif AngularDist.type == 'costheta':
        # Choose the altitude -- f(alt) = cos(alt)
        alt0 = AngularDist.altitude
        aa = (np.sin(alt0[0]), np.sin(alt0[1]))
        sinalt = np.random.rand(npackets) * (aa[1] - aa[0]) + aa[0]
        alt = np.arcsin(sinalt)

        # Choose the azimuth -- f(az) = 1/(azmax-azmin)
        az0, az1 = AngularDist.azimuth
        m = (az0, az1) if az0 < az1 else (az1, az0+2*np.pi)
        az = (m[0] + (m[1]-m[0])*np.random.rand(npackets)) % (2*np.pi*u.rad)

        # alt0 = AngularDist.altitude
        # aa = (np.sin(alt0[0]).value, np.sin(alt0[1]).value)
        # sinalt = np.linspace(aa[0], aa[1], 1001)
        # f_sinalt = sinalt**AngularDist.n
        # salt = mathMB.random_deviates_1d(sinalt, f_sinalt, npackets)
        # alt = np.arcsin(salt)
        #
        # # Choose the azimuth -- f(az) = 1/(azmax-azmin)
        # az0, az1 = AngularDist.azimuth
        # m = (az0, az1) if az0 < az1 else (az1, az0+2*np.pi)
        # az = (m[0] + (m[1]-m[0])*np.random.rand(npackets)) % (2*np.pi*u.rad)
    else:
        assert 0, 'Angular Distribution not defined.'

    # Find the velocity components in coordinate system centered on packet
    v_rad = np.sin(alt)                 # Radial component of velocity
    v_tan0 = np.cos(alt) * np.cos(az)   # Component along latitude (points E)
    v_tan1 = np.cos(alt) * np.sin(az)   # Component along longitude (points N)

    #######
    # Now rotate to proper surface point
    # v_ren = M # v_xyz => v_xyz = invert(M) # v_ren
    rr = np.sqrt(X0[0,:]**2 + X0[1,:]**2 + X0[2,:]**2)
    x0, y0, z0 = X0[0,:]/rr, X0[1,:]/rr, X0[2,:]/rr

    rad = np.array([x0, y0, z0])
    east = np.array([y0, -x0, np.zeros_like(z0)])
    north = np.array([-z0*x0, -z0*y0, x0**2+y0**2])

    east /= np.linalg.norm(east, axis=0)
    north /= np.linalg.norm(north, axis=0)

    v0 = v_tan0*north + v_tan1*east + v_rad*rad
    vx0 = v0[0,:] * vv
    vy0 = v0[1,:] * vv
    vz0 = v0[2,:] * vv
    V0 = np.array([vx0, vy0, vz0])

    return V0



'''Classes used by the Inputs class'''
import numpy as np
import pandas as pd
from astropy.time import Time
import astropy.units as u
from solarsystemMB import SSObject
from .database_connect import database_connect


def isNone(x):
    try:
        q = x.value
    except:
        q = x

    if type(q) == str:
        return f'is NULL' if q is None else f"= '{q}'"
    else:
        return f'is NULL' if q is None else f"= {q}"


def inRange(field, x, delta):
    return f'ABS({field} - {x}) <= {delta/2}'

dtor = np.pi/180.


class Geometry:
    def __init__(self, gparam):
        '''Geometry object: object to model
           Fields:
               planet
               StartPoint
               objects
               starttime
               phi
               subsolarpt = (subsolarlong, subsolarlat)
               TAA'''

        # Choose the planet
        if 'planet' in gparam:
            planet = gparam['planet'].title()
            self.planet = SSObject(planet)
        else:
            assert 0, 'Planet not defined.'

        objlist = [self.planet.object]
        if self.planet.moons is not None:
            objlist.extend([m.object for m in self.planet.moons])

        # Choose the starting point
        self.startpoint = (gparam['startpoint'].title()
                           if 'startpoint' in gparam
                           else self.planet.object)
        assert self.startpoint in objlist, 'Not a valid starting point'

        # Choose which objects to include
        # This is given as a list of names
        # Default = geometry.planet and geometry.startpoint
        if 'objects' in gparam:
            inc = set(i.strip().title()
                      for i in gparam['objects'].split(','))
        else:
            inc = set((self.planet.object, self.startpoint))

        for i in inc:
            assert i in objlist, 'Invalid object included: {}'.format(i)
        self.objects = set(SSObject(o) for o in inc)

        # Check to see if a starting time is given
        if 'time' in gparam:
            try:
                self.time = Time(gparam['time'].upper())
            except:
                assert 0, 'Time is not given in a valid format'

            assert 0, 'Need to figure out how to calculate orbital positions'
        else:
            # Initial positions are given
            self.time = None
            if 'phi' in gparam:
                phi = tuple(float(p)*u.rad for p in gparam['phi'].split(','))
                assert 0, 'Need to figure out best way to do this'
            elif len(self.objects) == 1:
                # No moons, so this isn't needed
                self.phi = [0.*u.rad]
            else:
                assert 0, ('Need to give either an observation time'
                           'or orbital position.')

        # Subsolar longitude and latitude
        subslong = (float(gparam['subsolarlong'])*u.rad if
                    'subsolarlong' in gparam else 0.*u.rad)
        subslat= (float(gparam['subsolarlat'])*u.rad if
                  'subsolarlat' in gparam else 0.*u.rad)
        self.subsolarpoint = (subslong, subslat)

        # True Anomaly Angle
        self.taa = float(gparam['taa']) if 'taa' in gparam else 0.
        self.taa *= u.rad

    def __str__(self):
        result = f'geometry.planet = {self.planet.object}\n'
        result += f'geometry.StartPoint = {self.startpoint}\n'
        oo = [o.object for o in self.objects]
        obs = ', '.join(oo)
        result += f'geometry.objects = {obs}'
        if self.time is not None:
            result += f'geometry.starttime = {self.time.iso}\n'
        else:
            result += f'geometry.startime not specified\n'
        if len(self.phi) != 0:
            result += 'geometry.phi XXX\n'
        result += 'geometry.subsolarpoint = ({}, {})\n'.format(*self.subsolarpoint)
        result += f'geometry.TAA = {self.taa}\n'
        return result

    def search(self, startlist=None):
        # Make list of objects in planet system
        objs = [obj.object for obj in self.objects]
        objs.sort()
        objs2 = ','.join(objs)

        if startlist is None:
            startstr = ''
        else:
            start_ = [str(s) for s in startlist]
            startstr = f"and geo_idnum in ({', '.join(start_)})"

        if self.time is None:
            # Fields to query:
            #   planet, startpoint, objects, phi, subsolarpoint, TAA
            dtaa = (5.*u.deg).to(u.rad)
            taa = [self.taa-dtaa/2., self.taa+dtaa/2.]
            taa = [taa[0].value, taa[1].value]
            if taa[0] < 0.:
                taabit = '(taa>={} or taa<{})'.format(2*np.pi+taa[0], taa[1])
            elif taa[1] > 2*np.pi:
                taabit = '(taa>={} or taa<{})'.format(taa[0],
                                                      taa[1] % (2*np.pi))
            else:
                taabit = inRange('taa', self.taa.value, dtaa.value)

            phi = [p.value for p in self.phi]
            assert phi[0] == 0., 'phi for planet should be zero.'
            ptxt = [inRange('phi[{}]'.format(i+1), p, 5.*dtor) for
                    i,p in enumerate(phi)]
            ptxt2 = ' and '.join(ptxt)

            sspt0 = inRange('subsolarpt[0]', self.subsolarpoint[0].value,
                            5*dtor)
            sspt1 = inRange('subsolarpt[1]', self.subsolarpoint[1].value,
                            5*dtor)

            query = f'''SELECT geo_idnum FROM geometry
                        WHERE planet = '{self.planet.object}' and
                              startpoint = '{self.startpoint}' and
                              objects = ARRAY['{objs2}']::SSObject[] and
                              {ptxt2} and
                              {sspt0} and
                              {sspt1} and
                              {taabit} {startstr}'''
        else:
            # Fields to query
            # planet, StartPoint, objects, time
            # query =
            assert 0, 'Not working yet.'

        with database_connect() as con:
            result = pd.read_sql(query, con)
        if len(result) == 0:
            return None
        else:
            return result.geo_idnum.to_list()
###############################################################


class StickingInfo:
    '''
    stickcoef
    tsurf
    stickfn
    stick_mapfile
    epsilon
    n
    tmin
    emitfn
    accom_mapfile
    accom_factor
    '''

    def __init__(self, sparam):
        self.stickcoef = (float(sparam['stickcoef'])
                          if 'stickcoef' in sparam
                          else 1.)
        if self.stickcoef > 1.:
            self.stickcoef = 1.

        # Defaults
        self.tsurf = None
        self.stickfn = sparam['stickfn'] if 'stickfn' in sparam else None
        self.stick_mapfile = None
        self.epsilon = None
        self.n = None
        self.tmin = None
        self.emitfn = sparam['emitfn'] if 'emitfn' in sparam else None
        self.accom_mapfile = None
        self.accom_factor = None

        # Set the stick function parameters
        if self.stickcoef == 1:
            # Complete sticking
            self.stickfn = 'complete'
        elif self.stickcoef > 0.:
            # Cosnstant stick function
            self.stickfn = 'constant'
        elif (self.stickcoef == -1) and (self.stickfn == 'use_map'):
            self.stick_mapfile = sparam['stick_mapfile']
        elif (self.stickcoef == -1) and (self.stickfn == 'linear'):
            self.epsilon = float(sparam['epsilon'])
            self.n = float(sparam['n']) if 'n' in sparam else 1.
            self.tmin = float(sparam['tmin'])*u.K
        elif (self.stickcoef == -1) and (self.stickfn == 'cossza'):
            self.n = float(sparam['n']) if 'n' in sparam else 1.
            self.tmin = float(sparam['tmin'])*u.K
        else:
            assert 0, 'sticking_info.stickfn not given or invalid.'

        # set the re-emission function parameters
        if self.emitfn == 'use_map':
            self.accom_mapfile = sparam['accom_mapfile']
        elif self.emitfn == 'maxwellian':
            ac = float(sparam['accom_factor'])
            if ac < 0:
                ac = 0.
            elif ac > 1:
                ac = 1.
            else:
                pass
            self.accom_factor = ac
        elif self.emitfn == 'elastic scattering':
            pass
        else:
            pass

    def __str__(self):
        result = f'sticking_info.stickcoef = {self.stickcoef}\n'
        if self.stickfn is not None:
            result += f'sticking_info.stickfn = {self.stickfn}\n'
        if self.tsurf is not None:
            result += f'sticking_info.tsurf = {self.tsurf}\n'
        if self.stick_mapfile is not None:
            result += f'sticking_info.stick_mapfile = {self.stick_mapfile}\n'
        if self.epsilon is not None:
            result += f'sticking_info.epsilon = {self.epsilon}\n'
        if self.n is not None:
            result += f'sticking_info.n = {self.n}\n'
        if self.tmin is not None:
            result += f'sticking_info.tmin = {self.tmin}\n'
        if self.emitfn is not None:
            result += f'sticking_info.emitfn = {self.emitfn}\n'
        if self.accom_mapfile is not None:
            result += f'sticking_info.accom_mapfile = {self.accom_mapfile}\n'
        if self.accom_factor is not None:
            result += f'sticking_info.accom_factor = {self.accom_factor}\n'
        return result

    def search(self, startlist=None):
        if startlist is None:
            startstr = ''
        else:
            start_ = [str(s) for s in startlist]
            startstr = f"and st_idnum in ({', '.join(start_)})"

        if self.stickcoef == 1:
            query = f'''SELECT st_idnum FROM sticking_info
                        WHERE stickcoef=1 {startstr}'''
        else:
            query = f'''SELECT st_idnum FROM sticking_info
                        WHERE stickcoef={self.stickcoef}
                             tsurf {self.tsurf} and
                             stickfn {self.stickfn} and
                             stick_mapfile {self.stick_mapfile} and
                             epsilon {self.epsilon} and
                             n {self.n} and
                             tmin {self.tmin} and
                             emitfn {self.emitfn} and
                             accom_mapfile {self.accom_mapfile}
                             {startstr}'''

        with database_connect() as con:
            result = pd.read_sql(query, con)
        if len(result) == 0:
            return None
        else:
            return result.st_idnum.to_list()


class Forces:
    def __init__(self, fparam):
        '''
        gravity
        radpres
        '''
        self.gravity = (bool(int(float(fparam['gravity'])))
                        if 'gravity' in fparam
                        else False)
        self.radpres = (bool(int(float(fparam['radpres'])))
                        if 'radpres' in fparam
                        else False)

    def __str__(self):
        result = f'forces.gravity = {self.gravity}\n'
        result += f'forces.radpres = {self.radpres}\n'
        return result

    def search(self, startlist=None):
        if startlist is None:
            startstr = ''
        else:
            start_ = [str(s) for s in startlist]
            startstr = f"and f_idnum in ({', '.join(start_)})"

        query = f'''SELECT f_idnum FROM forces
                    WHERE gravity={self.gravity} and
                          radpres={self.radpres} {startstr}'''
        with database_connect() as con:
            result = pd.read_sql(query, con)

        if len(result) == 0:
            return None
        else:
            return result.f_idnum.to_list()

###############################################################


class SpatialDist:
    def __init__(self, sparam):
        '''
        type
        exobase
        use_map
        mapfile
        lonrange
        latrange
        '''

        # Set defaults
        self.type = sparam['type']
        self.exobase = 0.
        self.use_map = False
        self.mapfile = None
        self.longitude = None
        self.latitude = None

        if self.type == 'surface':
            self.exobase = (float(sparam['exobase'])
                            if 'exobase' in sparam
                            else 1.)  # Unit gets set later
            self.use_map = (bool(int(sparam['use_map']))
                            if 'use_map' in sparam
                            else False)
            if self.use_map:
                self.mapfile = sparam['mapfile']

            long0 = (float(sparam['longitude0'])*u.rad
                if 'longitude0' in sparam else 0.*u.rad)
            long1 = (float(sparam['longitude1'])*u.rad
                if 'longitude1' in sparam else 2*np.pi*u.rad)
            lat0 = (float(sparam['latitude0'])*u.rad
                if 'latitude0' in sparam else -np.pi/2.*u.rad)
            lat1 = (float(sparam['latitude1'])*u.rad
                if 'latitude1' in sparam else np.pi/2.*u.rad)
            self.longitude = (long0, long1)
            self.latitude = (lat0, lat1)
        elif self.type == 'surfacespot':
            self.exobase = (float(sparam['exobase'])
                            if 'exobase' in sparam
                            else 1.)  # Unit gets set later
            lon = (float(sparam['longitude'])*u.rad
                   if 'longitude' in sparam else 0.*u.rad)
            lat = (float(sparam['latitude'])*u.rad
                   if 'latitude' in sparam else 0*u.rad)
            sigma = (float(sparam['sigma'])*u.rad
                     if 'sigma' in sparam else 25*u.deg)
            if sigma < 0*u.deg:
                sigma = 0*u.deg
            elif sigma > 90*u.deg:
                sigma = 90*u.deg
            else:
                pass

            self.longitude = (lon, sigma.to(u.rad))
            self.latitude = (lat, sigma.to(u.rad))
            # self.sigma = sigma
        elif self.type== 'idlversion':
            if 'idlinputfile' in sparam:
                self.mapfile = sparam['idlinputfile']
            else:
                assert 0, 'Must specify idlinputfile'

        else:
            assert 0, f'{self.type} distribution not defined yet.'

    def __str__(self):
        result = f'spatialdist.type = {self.type}\n'
        result += f'spatialdist.exobase = {self.exobase}\n'
        result += f'spatialdist.use_map = {self.use_map}\n'
        result += f'spatialdist.mapfile = {self.mapfile}\n'
        if self.longitude is not None:
            result += 'spatialdist.longitude = ({:0.2f}, {:0.2f})\n'.format(*self.longitude)
        if self.latitude is not None:
            result += 'spatialdist.latitude = ({:0.2f}, {:0.2f})\n'.format(*self.latitude)
        return result

    def search(self, startlist=None):
        if startlist is None:
            startstr = ''
        else:
            start_ = [str(s) for s in startlist]
            startstr = f"and spat_idnum in ({', '.join(start_)})"

        if self.longitude is None:
            long0 = 'longitude[1] = 0.'
            long1 = 'longitude[2] = 0.'
        else:
            long0 = inRange('longitude[1]', self.longitude[0].value, 5*dtor)
            long1 = inRange('longitude[2]', self.longitude[1].value, 5*dtor)

        if self.latitude is None:
            lat0 = 'latitude[1] = 0.'
            lat1 = 'latitude[2] = 0.'
        else:
            lat0 = inRange('latitude[1]', self.latitude[0].value, 5*dtor)
            lat1 = inRange('latitude[2]', self.latitude[1].value, 5*dtor)

        query = f'''SELECT spat_idnum FROM spatialdist
                    WHERE type = '{self.type}' and
                         {inRange('exobase', self.exobase, 0.05)} and
                         use_map {isNone(self.use_map)} and
                         mapfile {isNone(self.mapfile)} and
                         {long0} and
                         {long1} and
                         {lat0} and
                         {lat1} {startstr}'''

        with database_connect() as con:
            result = pd.read_sql(query, con)
        if len(result) == 0:
            return None
        else:
            return result.spat_idnum.to_list()
###############################################################


class SpeedDist:
    '''
    type
    vprob
    sigma
    U
    alpha
    beta
    temperature
    delv
    '''


    def __init__(self, sparam):
        self.type = sparam['type']

        # Defaults
        self.vprob = None
        self.sigma = None
        self.U = None
        self.alpha = None
        self.beta = None
        self.temperature = None
        self.delv = None

        if self.type == 'gaussian':
            self.vprob = float(sparam['vprob'])*u.km/u.s
            self.sigma = float(sparam['sigma'])*u.km/u.s
        elif self.type == 'sputtering':
            self.U = float(sparam['u'])*u.eV
            self.alpha = float(sparam['alpha'])
            self.beta = float(sparam['beta'])
        elif self.type == 'maxwellian':
            self.temperature = float(sparam['temperature'])*u.K
        elif self.type == 'flat':
            self.vprob = float(sparam['vprob'])*u.km/u.s
            self.delv = float(sparam['delv'])*u.km/u.s
        else:
            assert 0, f'SpeedDist.type = {self.type} not available'

    def __str__(self):
        result = f'SpeedDist.type = {self.type}\n'
        if self.vprob is not None:
            result += f'SpeedDist.vprob = {self.vprob}\n'
        if self.sigma is not None:
            result += f'SpeedDist.sigma = {self.sigma}\n'
        if self.U is not None:
            result += f'SpeedDist.U = {self.U}\n'
        if self.alpha is not None:
            result += f'SpeedDist.alpha = {self.alpha}\n'
        if self.beta is not None:
            result += f'SpeedDist.beta = {self.beta}\n'
        if self.temperature is not None:
            result += f'SpeedDist.temperature = {self.temperature}\n'
        if self.delv is not None:
            result += f'SpeedDist.delv = {self.delv}\n'

        return result

    def search(self, startlist=None):
        if startlist is None:
            startstr = ''
        else:
            start_ = [str(s) for s in startlist]
            startstr = f"and spd_idnum in ({', '.join(start_)})"

        if self.vprob is None:
            vstr = 'vprob is NULL'
        else:
            vstr = inRange('vprob', self.vprob.value,
                           self.vprob.value*0.05)

        if self.temperature is None:
            Tstr = 'temperature is NULL'
        else:
            Tstr = inRange('temperature', self.temperature.value,
                           self.temperature.value*0.05)

        query = f'''SELECT spd_idnum FROM speeddist
                    WHERE type = '{self.type}' and
                          {vstr} and
                          sigma {isNone(self.sigma)} and
                          U  {isNone(self.U)} and
                          alpha  {isNone(self.alpha)} and
                          beta  {isNone(self.beta)} and
                          {Tstr} and
                          delv  {isNone(self.delv)} {startstr}'''

        with database_connect() as con:
            result = pd.read_sql(query, con)
        if len(result) == 0:
            return None
        else:
            return result.spd_idnum.to_list()


class AngularDist:
    '''
    type
    azimuth
    altitude
    n
    '''

    def __init__(self, aparam, spatialdist):
        self.type = aparam['type'] if 'type' in aparam else None
        self.azimuth = None
        self.altitude = None
        self.n = None

        if self.type is None:
            pass
        elif self.type == 'radial':
            pass
        elif self.type == 'isotropic':
            if 'azimuth' in aparam:
                self.azimuth = tuple(float(a)*u.rad
                    for a in aparam['azimuth'].split(','))
                assert len(self.azimuth) == 2, (
                    'AngularDist.azimuth must have two values.')
            else:
                self.azimuth = (0*u.rad, 2*np.pi*u.rad)

            if 'altitude' in aparam:
                self.altitude = tuple(float(a)*u.rad
                    for a in aparam['altitude'].split(','))
                assert len(self.altitude) ==2, (
                    'AngularDist.altitude must have two values.')
            else:
                altmin = (0.*u.rad if 'surface' in spatialdist.type
                          else -np.pi/2.*u.rad)
                self.altitude = (altmin, np.pi/2.*u.rad)
        elif self.type == 'costheta':
            if 'azimuth' in aparam:
                self.azimuth = tuple(float(a)*u.rad
                    for a in aparam['azimuth'].split(','))
                assert len(self.azimuth) == 2, (
                    'AngularDist.azimuth must have two values.')
            else:
                self.azimuth = (0*u.rad, 2*np.pi*u.rad)

            if 'altitude' in aparam:
                self.altitude = tuple(float(a)*u.rad
                    for a in aparam['altitude'].split(','))
                assert len(self.altitude) ==2, (
                    'AngularDist.altitude must have two values.')
            else:
                altmin = (0.*u.rad if 'surface' in spatialdist.type
                          else -np.pi/2.*u.rad)
                self.altitude = (altmin, np.pi/2.*u.rad)

            self.n = float(aparam['n']) if 'n' in aparam else 1.

    def __str__(self):
        result = f'AngularDist.type = {self.type}\n'
        if self.altitude is not None:
            result += 'AngularDist.altitude = ({:0.2f}, {:0.2f})\n'.format(*self.altitude)
        if self.azimuth is not None:
            result += 'AngularDist.azimuth = ({:0.2f}, {:0.2f})\n'.format(*self.azimuth)
        if self.n is not None:
            result += f'AngularDist.n = {self.n}\n'

        return result

    def search(self, startlist=None):
        if startlist is None:
            startstr = ''
        else:
            start_ = [str(s) for s in startlist]
            startstr = f"and ang_idnum in ({', '.join(start_)})"

        if self.azimuth is None:
            az0 = 'azimuth[1] is NULL'
            az1 = 'azimuth[2] is NULL'
        else:
            az0 = inRange('azimuth[1]', self.azimuth[0].value, 5*dtor)
            az1 = inRange('azimuth[2]', self.azimuth[1].value, 5*dtor)

        if self.altitude is None:
            alt0 = 'altitude[1] is NULL'
            alt1 = 'altitude[2] is NULL'
        else:
            alt0 = inRange('altitude[1]', self.altitude[0].value, 5*dtor)
            alt1 = inRange('altitude[2]', self.altitude[1].value, 5*dtor)
        n = isNone(self.n)

        query = f'''SELECT ang_idnum from angulardist
                    WHERE type = '{self.type}' and
                          {az0} and {az1} and
                          {alt0} and {alt1} and
                          n {n} {startstr}'''
        with database_connect() as con:
            result = pd.read_sql(query, con)

        if len(result) == 0:
            return None
        else:
            return result.ang_idnum.to_list()
###############################################################


class Options:
    '''
    endtime
    resolution
    at_once
    atom
    lifetime
    fullsystem
    outeredge
    motion
    streamlines
    nsteps
    '''

    def __init__(self, oparam, planet):
        self.endtime = float(oparam['endtime'])*u.s
        self.at_once = (bool(int(oparam['at_once'])) if 'at_once'
                        in oparam else False)
        self.atom = oparam['atom'].title()
        self.motion = (bool(int(oparam['motion'])) if 'motion'
                       in oparam else True)
        self.lifetime = (float(oparam['lifetime'])*u.s
                         if 'lifetime' in oparam else 0.*u.s)

        if 'fullsystem' in oparam:
            self.fullsystem = bool(int(oparam['fullsystem']))
        else:
            self.fullsystem = False if planet == 'Mercury' else True

        if not(self.fullsystem):
            self.outeredge = (float(oparam['outeredge'])
                              if 'outeredge' in oparam else 20.)
            # Units added later
        else:
            self.outeredge = None

        self.streamlines = (bool(int(oparam['streamlines']))
                            if 'streamlines' in oparam else False)
        if self.streamlines:
            self.nsteps = (int(oparam['nsteps'])
                           if 'nsteps' in oparam else 1000)
            self.resolution = None
        else:
            self.nsteps = None
            self.resolution = (float(oparam['resolution'])
                               if 'resolution' in oparam else 1e-3)

    def __str__(self):
        result = f'options.endtime = {self.endtime}\n'
        result += f'options.resolution = {self.resolution}\n'
        result+= f'options.at_once = {self.at_once}\n'
        result+= f'options.atom = {self.atom}\n'
        result+= f'options.motion = {self.motion}\n'
        result+= f'options.lifetime = {self.lifetime}\n'
        result+= f'options.fullsystem = {self.fullsystem}\n'
        if self.outeredge is not None:
            result+= f'options.outeredge = {self.outeredge}\n'
        result+= f'options.streamlines = {self.streamlines}\n'
        if self.nsteps is not None:
            result+= f'options.nsteps = {self.nsteps}'

        return result

    def search(self, startlist=None):
        if startlist is None:
            startstr = ''
        else:
            start_ = [str(s) for s in startlist]
            startstr = f"and opt_idnum in ({', '.join(start_)})"

        endtime = inRange('endtime', self.endtime.value,
                          self.endtime.value*0.05)
        outeredge = isNone(self.outeredge)
        nsteps = isNone(self.nsteps)
        res = isNone(self.resolution)

        query = f'''SELECT opt_idnum from options
                    WHERE {endtime} and
                          resolution {res} and
                          at_once = {self.at_once} and
                          atom = '{self.atom}' and
                          lifetime = {self.lifetime.value} and
                          fullsystem = {self.fullsystem} and
                          outeredge {outeredge} and
                          motion = {self.motion} and
                          streamlines = {self.streamlines} and
                          nsteps {nsteps} {startstr}'''
        with database_connect() as con:
            result = pd.read_sql(query, con)

        if len(result) == 0:
            return None
        else:
            return result.opt_idnum.to_list()



import os.path
import numpy as np
import numpy.matlib as npmat
import pandas as pd
import pickle
import astropy.units as u
from solarsystemMB import SSObject
from mathMB import rotation_matrix
from .ModelResults import (ModelResult, read_format, results_loadfile,
                           results_packet_weighting)
from .database_connect import database_connect

quantities = ['column', 'intensity', 'density']

def image_rotation(image):
    slong = image.subobslongitude
    slat = image.subobslatitude

    pSun = np.array([0., -1., 0.])
    pObs = np.array([np.sin(slong)*np.cos(slat),
                     -np.cos(slong)*np.cos(slat),
                     np.sin(slat)])
    if np.array_equal(pSun, pObs):
        M = npmat.identity(3)
    else:
        costh = np.dot(pSun, pObs)/np.linalg.norm(pSun)/np.linalg.norm(pObs)
        theta = np.arccos(np.clip(costh, -1, 1))
        axis = np.cross(pSun, pObs)
        M = rotation_matrix(theta, axis)

    #M = np.transpose(M)
    return M


class ModelImage(ModelResult):
    def __init__(self, inputs, formatfile, filenames=None):
        self.type = 'image'

        # Read in the format file
        if isinstance(formatfile, str):
            format_ = read_format(formatfile)

            quantities = ['column', 'radiance']
            if format_['quantity'] in quantities:
                self.quantity = format_['quantity']
            else:
                assert 0, 'Quantity not specified'

            if format_['quantity'] == 'radiance':
                self.mechanism = tuple(m.strip()
                                  for m in format_['mechanism'].split(','))
                if 'wavelength' in format_:
                    self.wavelength = tuple(int(m.strip())*u.AA
                                 for m in format_['wavelength'].split(','))
                elif inputs.options.atom == 'Na':
                    self.wavelength = (5891*u.AA, 5897*u.AA)
                elif inputs.options.atom == 'Ca':
                    self.wavelength = (4227*u.AA,)
                elif inputs.options.atom == 'Mg':
                    self.wavelength = (2852*u.AA,)
                else:
                    assert 0, f'No default wavelength for {input.options.atom}'
            else:
                pass

            self.origin = SSObject(format_['origin'])
            self.unit = u.def_unit('R_' + self.origin.object,
                                   self.origin.radius)

            dimtemp = format_['dims'].split(',')
            self.dims = np.array((int(dimtemp[0]), int(dimtemp[1])))

            centtemp = format_['center'].split(',')
            self.center = np.array((float(centtemp[0]), float(centtemp[1])))
            self.center *= self.unit

            widtemp = format_['width'].split(',')
            self.width = np.array((float(widtemp[0]), float(widtemp[1])))
            self.width *= self.unit

            self.subobslongitude = float(format_['subobslongitude'])*u.rad
            self.subobslatitude = float(format_['subobslatitude'])*u.rad
        elif isinstance(formatfile, dict):
            assert 0, 'Not working right yet'
            format_ = formatfile
            if format_['quantity'] in quantities:
                self.quantity= format_['quantity']
            else:
                assert 0, 'Quantity not specified'

            self.origin = SSObject(format_['origin'])
            unit = u.def_unit('R_' + self.origin.object, self.origin.radius)

            self.dims = format_['dims']
            self.center = format_['center']
            self.width = format_['width']

            if isinstance(format_['subobslongitude'][0], (int, float)):
                self.subobslongitude = (format_['subobslongitude'][0]*u.rad,
                                        format_['subobslongitude'][1]*u.rad)
            elif isinstance(format_['subobslongitude'][0], type(1*u.rad)):
                self.subobslongitude = format_['subobslongitude']

            if isinstance(format_['subobslatitude'][0], (int, float)):
                self.subobslatitude = (format_['subobslatitude'][0]*u.rad,
                                        format_['subobslatitude'][1]*u.rad)
            elif isinstance(format_['subobslatitude'][0], type(1*u.rad)):
                self.subobslatitude = format_['subobslatitude']

        else:
            assert 0, 'Format problem.'

        # Set up universal result stuff
        ModelResult.__init__(self, inputs)
        if isinstance(filenames, str):
            print('Setting filenames breaks calibration.')
            self.filenames = [filenames]
        elif isinstance(filenames, list):
            print('Setting filenames breaks calibration.')
            self.filenames = filenames
        else:
            pass

        self.image = np.zeros(self.dims)
        self.packet_image = np.zeros(self.dims)
        self.blimits = None
        immin = self.center - self.width/2.
        immax = self.center + self.width/2.
        scale = self.width/(self.dims-1) # Rplan/pix
        self.Apix = (scale[0]*scale[1]).to(u.cm**2)

        self.xaxis = np.linspace(immin[0], immax[0], self.dims[0])
        self.zaxis = np.linspace(immin[1], immax[1], self.dims[1])

        for i, fname in enumerate(self.filenames):
            # Search to see if its already been done
            image_, packets_ = self.restore(fname)

            if image_ is None:
                image_, packets_ = self.create_image(fname)
                print(f'Completed image {i+1} of {len(self.filenames)}')
            else:
                print(f'Image {i+1} of {len(self.filenames)} '
                       'previously completed.')

            self.image += image_
            self.packet_image += packets_

        self.image = self.image * self.atoms_per_packet

    def save(self, fname, image, packets):
        con = database_connect()
        cur = con.cursor()

        # Determine the id of the outputfile
        idnum_ = pd.read_sql(f'''SELECT idnum
                                FROM outputfile
                                WHERE filename='{fname}' ''', con)
        idnum = int(idnum_.idnum[0])

        # Insert the image into the database
        if self.quantity == 'radiance':
            mech = ', '.join(sorted([m for m in self.mechanism]))
            wave_ = sorted([w.value for w in self.wavelength])
            wave = ', '.join([str(w) for w in wave_])
        else:
            mech = None
            wave = None

        dims = f'ARRAY[{self.dims[0]}, {self.dims[1]}]'
        center = f'ARRAY[{self.center[0].value}, {self.center[1].value}]'
        width = f'ARRAY[{self.width[0].value}, {self.width[1].value}]'

        cur.execute(f'''INSERT into modelimages (out_idnum, quantity,
                            origin, dims, center, width, subobslongitude,
                            subobslatitude, mechanism, wavelength)
                        values (%s, %s, %s, {dims}, {center}, {width},
                            %s, %s, %s, %s)''',
                    (idnum, self.quantity, self.origin.object,
                    self.subobslongitude.value, self.subobslatitude.value,
                    mech, wave))

        # Determine the savefile name
        idnum_ = pd.read_sql(f'''SELECT idnum
                                FROM modelimages
                                WHERE filename is NULL''', con)
        assert len(idnum_) == 1
        idnum = int(idnum_.idnum[0])

        savefile = os.path.join(os.path.dirname(fname), f'image.{idnum}.pkl')
        with open(savefile, 'wb') as f:
            pickle.dump((image, packets), f)
        cur.execute(f'''UPDATE modelimages
                        SET filename=%s
                        WHERE idnum = %s''', (savefile, idnum))
        con.close()

    def restore(self, fname):
        con = database_connect()

        # Determine the id of the outputfile
        idnum_ = pd.read_sql(f'''SELECT idnum
                                FROM outputfile
                                WHERE filename='{fname}' ''', con)
        oid = idnum_.idnum[0]

        if self.quantity == 'radiance':
            mech = ("mechanism = '" +
                    ", ".join(sorted([m for m in self.mechanism])) +
                    "'")
            wave_ = sorted([w.value for w in self.wavelength])
            wave = ("wavelength = '" +
                    ", ".join([str(w) for w in wave_]) +
                    "'")
        else:
            mech = 'mechanism is NULL'
            wave = 'wavelength is NULL'

        result = pd.read_sql(
            f'''SELECT filename FROM modelimages
                WHERE out_idnum = {oid} and
                      quantity = '{self.quantity}' and
                      origin = '{self.origin.object}' and
                      dims[1] = {self.dims[0]} and
                      dims[2] = {self.dims[1]} and
                      center[1] = {self.center[0].value} and
                      center[2] = {self.center[1].value} and
                      width[1] = {self.width[0].value} and
                      width[2] = {self.width[1].value} and
                      subobslongitude = {self.subobslongitude.value} and
                      subobslatitude = {self.subobslatitude.value} and
                      {mech} and
                      {wave}''', con)

        assert len(result) <= 1
        if len(result) == 1:
            savefile = result.filename[0]
            image, packets = pickle.load(open(savefile, 'rb'))
        else:
            image, packets = None, None

        con.close()

        return image, packets

    def create_image(self, fname):
        # Determine the proper frame rotation
        M = image_rotation(self)

        # Load data in solar reference frame
        output = results_loadfile(fname)
        radvel_sun = output.vy + output.vrplanet

        # packet positions in an array
        pts_sun = np.array((output.x, output.y, output.z))*output.x.unit
        frac = output.frac

        # Rotate to observer frame
        pts_obs = np.matmul(M, pts_sun)

        # Determine which packets are not blocked by planet
        rhosqr_obs = pts_obs[0,:]**2 + pts_obs[2,:]**2
        inview = (rhosqr_obs.value > 1) | (pts_obs[1,:].value < 0)
        frac *= inview

        # Which packets are in sunlight
        rhosqr_sun = pts_sun[0,:]**2 + pts_sun[2,:]**2
        out_of_shadow = (rhosqr_sun.value > 1) | (pts_sun[1,:].value < 0)

        # Packet weighting
        weight = results_packet_weighting(self, radvel_sun, frac,
                                  out_of_shadow, output.aplanet) / self.Apix
        assert pts_obs.unit == self.xaxis.unit

        dx = (self.xaxis[1]-self.xaxis[0])/2.
        bx = np.append(self.xaxis.value-dx.value,
                       self.xaxis[-1].value+dx.value)*self.xaxis.unit
        dz = (self.zaxis[1]-self.zaxis[0])/2.
        bz = np.append(self.zaxis.value-dz.value,
                       self.zaxis[-1].value+dz.value)*self.zaxis.unit

        image, _, _ = np.histogram2d(pts_obs[0,:], pts_obs[2,:],
                                  weights=weight, bins=(bx, bz))
        packets, _, _ = np.histogram2d(pts_obs[0,:], pts_obs[2,:],
                                  bins=(bx, bz))

        self.save(fname, image, packets)
        del output

        return image, packets

    def display(self, savefile='image.png', limits=None, show=True):
        import matplotlib.pyplot as plt
        import matplotlib.image as mpimg
        from matplotlib.colors import LogNorm
        from astropy.visualization import (PercentileInterval, LogStretch,
                                           ImageNormalize)

        extent = (np.min(self.xaxis.value), np.max(self.xaxis.value),
                  np.min(self.zaxis.value), np.max(self.zaxis.value))
        if self.unit.__str__() == 'R_Mercury':
            ustr = '$R_{M}$'
        else:
            ustr = '$R_{obj}$'

        # Determine limits if none given
        if limits is None:
            interval = PercentileInterval(95)
            self.blimits = interval.get_limits(self.image[self.image > 0])
        elif len(limits) == 2:
            self.blimits = limits
        else:
            assert 0, 'Problem with the display limits'

#        norm = ImageNormalize(self.image, stretch=LogStretch(),
#                              vmin=self.blimits[0], vmax=self.blimits[1])

        # Make the figure
        fig, ax = plt.subplots(figsize=(12,12))
#        im = ax.imshow(self.image, cmap='afmhot', extent=extent,
#                       norm=LogNorm(vmin=self.blimits[0], vmax=self.blimits[1]))
        im = ax.imshow(self.image.transpose(), cmap='afmhot', extent=extent,
                       norm=LogNorm(vmin=self.blimits[0], vmax=self.blimits[1]))
        ax.set_xlabel(f'Distance ({ustr})')
        ax.set_ylabel(f'Distance ({ustr})')
        ax.set_title(f'{self.inputs.options.atom} {self.quantity.title()}')

        # Make the colorbar
        if self.quantity == 'column':
            clabel = f'$N_{{ {self.inputs.options.atom} }}\ cm^{{-2}}$'
        else:
            clabel = f'$I_{{ {self.inputs.options.atom} }} R$'
        cbar = fig.colorbar(im, shrink=0.7, label=clabel)

        # Put Planet's disk in the middle
        xc, yc = (np.cos(np.linspace(0, 2*np.pi, 1000)),
                  np.sin(np.linspace(0, 2*np.pi, 1000)))
        ax.fill(xc, yc, 'y')

        # Display the plot if requested
        if show:
            plt.show()

        # Save the figure
        plt.savefig(savefile)
        plt.close()

        # Returns figure, axes, and colorbar for further work if desired
        return fig, ax, cbar



''' Computes acceleration and ionization on a packet due to specified forces

Gravitational acceleration

Equations of motion:
    dvxdt = sum_objects (GM * (x-x_obj))/(r_obj)^3
    dvydt = sum_objects (GM * (y-y_obj))/(r_obj)^3
    dvzdt = sum_objects (GM * (z-z_obj))/(r_obj)^3
        -- r_obj = sqrt( (x-x_obj)^2 + (y-y_obj)^2 + (z-z_obj)^2 )
    dndt = instantaneous change in density

Current version: Assumes there is only a planet -- does not do moons yet
'''

import numpy as np
from astropy.time import Time
#from .xyz_to_magcoord import xyz_to_magcoord

def State(t, x, v, output):
    # compute gravitational acceleration
    if output.inputs.forces.gravity:
        r3 = (x[0,:]**2 + x[1,:]**2 + x[2,:]**2)**1.5
        agrav = output.GM * x/r3
    else:
        agrav = np.zeros_like(x)

    # compute radiation acceleration
    arad = np.zeros_like(x)
    if output.inputs.forces.radpres:
        rho = x[0,:]**2 + x[2,:]**2
        out_of_shadow = (rho > 1) | (x[1,:] < 0)

        # radial velocity of each packet realtive to the Sun
        vv = v[1,:] + output.vrplanet

        # Compute radiation acceleration
        arad[1,:] = (np.interp(vv, output.radpres.velocity,
                               output.radpres.accel) * out_of_shadow)
    else:
        pass

    # Compute total acceleration
    accel = agrav + arad
    assert np.all(np.isfinite(accel))

    # Compute ionization rate
    if output.inputs.options.lifetime > 0:
        # Explicitly set lifetime
        ionizerate = np.ones_like(t)/output.inputs.options.lifetime.value
    else:
        if output.loss_info.photo is not None:
            # Compute photoionization rate
            rho = x[0,:]**2 + x[2,:]**2
            out_of_shadow = (rho > 1) | (x[1,:] < 0)
            photorate = output.loss_info.photo * out_of_shadow
        else:
            photorate = 0.

        '''
        magcoord = xyz_to_magcoord(t, x, output.inputs, output.planet)

        if output.loss_info.eimp:
            # Compute electron impact rate
            assert 0, 'Electron impacts not set up yet'
        else:
            eimprate = 0.

        if output.loss_info.chX:
            # Compute charge exchange rate
            assert 0, 'Charge exchange not set up yet'
        else:
            chxrate = 0.
        '''

    ionizerate = photorate  #+ eimprate + chxrate

    return accel, ionizerate



from .Input import Input
from .Output import Output
from .modeldriver import modeldriver
from .LOSResult import LOSResult
from .produce_image import ModelImage
from .configure_model import configure_model
from .IDLout import IDLout


name = 'nexoclom'
__author__ = 'Matthew Burger'
__email__ = 'mburger@stsci.edu'
__version__ = '1.2.2'
database = 'thesolarsystemmb'


configure_model(force=False)
